# -*- coding: utf-8 -*-
"""
    Sellar discipline 2
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from .disc2_base import Disc2Base


class Disc2(Disc2Base):
    """ An OpenMDAO component to encapsulate Disc2 discipline """

    # pylint: disable=invalid-name
    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Evaluates the equation
        y2 = y1**(.5) + z1 + z2
        """

        z1 = inputs["z"][0]
        z2 = inputs["z"][1]
        y1 = inputs["y1"]

        # Note: this may cause some issues. However, y1 is constrained to be
        # above 3.16, so lets just let it converge, and the optimizer will
        # throw it out
        if y1.real < 0.0:
            y1 *= -1

        outputs["y2"] = y1 ** 0.5 + z1 + z2



"""
Module for managing OpenMDAO variables
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from copy import deepcopy
from typing import Dict, Hashable, List, Union

import numpy as np
import openmdao.api as om
import pandas as pd
from importlib_resources import open_text
from openmdao.core.system import System

from . import resources
from .utils import get_problem_after_setup, get_unconnected_input_names

# Logger for this module
_LOGGER = logging.getLogger(__name__)

DESCRIPTION_FILENAME = "variable_descriptions.txt"
# Metadata that will be ignore when checking variable equality and when adding variable
# to an OpenMDAO component
METADATA_TO_IGNORE = ["is_input", "tags", "size", "src_indices", "flat_src_indices", "distributed"]


class Variable(Hashable):
    """
    A class for storing data of OpenMDAO variables.

    Instantiation is expected to be done through keyword arguments only.

    Beside the mandatory parameter 'name, kwargs is expected to have keys
    'value', 'units' and 'desc', that are accessible respectively through
    properties :meth:`name`, :meth:`value`, :meth:`units` and :meth:`description`.

    Other keys are possible. They match the definition of OpenMDAO's method
    :meth:`Component.add_output` described
    `here <http://openmdao.org/twodocs/versions/latest/_srcdocs/packages/core/
    component.html#openmdao.core.component.Component.add_output>`_.

    These keys can be listed with class method :meth:`get_authorized_keys`.
    **Any other key in kwargs will be silently ignored.**

    Special behaviour: :meth:`description` will return the content of kwargs['desc']
    unless these 2 conditions are met:
     - kwargs['desc'] is None or 'desc' key is missing
     - a description exists in FAST-OAD internal data for the variable name
    Then, the internal description will be returned by :meth:`description`

    :param kwargs: the attributes of the variable, as keyword arguments
    """

    # Will store content of DESCRIPTION_FILE_PATH once and for all
    _variable_descriptions = {}

    # Default metadata
    _base_metadata = {}

    def __init__(self, name, **kwargs: Dict):
        super().__init__()

        self.name = name
        """ Name of the variable """

        self.metadata: Dict = {}
        """ Dictionary for metadata of the variable """

        # Initialize class attributes once at first instantiation -------------
        if not self._variable_descriptions:
            # Class attribute, but it's safer to initialize it at first instantiation
            with open_text(resources, DESCRIPTION_FILENAME) as desc_io:
                vars_descs = np.genfromtxt(desc_io, delimiter="\t", dtype=str)
            self.__class__._variable_descriptions.update(vars_descs)

        if not self._base_metadata:
            # Get variable base metadata from an ExplicitComponent
            comp = om.ExplicitComponent()
            # get attributes
            metadata = comp.add_output(name="a")

            self.__class__._base_metadata = metadata
            self.__class__._base_metadata["value"] = 1.0
            self.__class__._base_metadata["tags"] = set()
            self.__class__._base_metadata["shape"] = None
        # Done with class attributes ------------------------------------------

        self.metadata = self.__class__._base_metadata.copy()
        self.metadata.update(kwargs)
        self._set_default_shape()

        # If no description, add one from DESCRIPTION_FILE_PATH, if available
        if not self.description and self.name in self._variable_descriptions:
            self.description = self._variable_descriptions[self.name]

    @classmethod
    def get_openmdao_keys(cls):
        """

        :return: the keys that are used in OpenMDAO variables
        """
        # As _base_metadata is initialized at first instantiation, we create an instance to
        # ensure it has been done
        cls("dummy")

        return cls._base_metadata.keys()

    @property
    def value(self):
        """ value of the variable"""
        return self.metadata.get("value")

    @value.setter
    def value(self, value):
        self.metadata["value"] = value
        self._set_default_shape()

    @property
    def units(self):
        """ units associated to value (or None if not found) """
        return self.metadata.get("units")

    @units.setter
    def units(self, value):
        self.metadata["units"] = value

    @property
    def description(self):
        """ description of the variable (or None if not found) """
        return self.metadata.get("desc")

    @description.setter
    def description(self, value):
        self.metadata["desc"] = value

    @property
    def is_input(self):
        """ True if variable is a problem input, False if it is an output (None if information not found) """
        return self.metadata.get("is_input")

    @is_input.setter
    def is_input(self, value):
        self.metadata["is_input"] = value

    def _set_default_shape(self):
        """ Automatically sets shape if not set"""
        if self.metadata["shape"] is None:
            shape = np.shape(self.value)
            if not shape:
                shape = (1,)
            self.metadata["shape"] = shape

    def __eq__(self, other):
        # same arrays with nan are declared non equals, so we need a workaround
        my_metadata = dict(self.metadata)
        other_metadata = dict(other.metadata)
        my_value = np.asarray(my_metadata.pop("value"))
        other_value = np.asarray(other_metadata.pop("value"))

        # Let's also ignore unimportant keys
        for key in METADATA_TO_IGNORE:
            if key in my_metadata:
                del my_metadata[key]
            if key in other_metadata:
                del other_metadata[key]

        return (
            isinstance(other, Variable)
            and self.name == other.name
            and ((my_value == other_value) | (np.isnan(my_value) & np.isnan(other_value))).all()
            and my_metadata == other_metadata
        )

    def __repr__(self):
        return "Variable(name=%s, metadata=%s)" % (self.name, self.metadata)

    def __hash__(self) -> int:
        return hash("var=" + self.name)  # Name is normally unique


class VariableList(list):
    """
    Class for storing OpenMDAO variables

    A list of Variable instances, but items can also be accessed through variable names.

    There are 2 ways for adding a variable::

        # Assuming these Python variables are ready
        var_1 = Variable('var/1', value=0.)
        metadata_2 = {'value': 1., 'units': 'm'}

        # ... a VariableList instance can be populated like this
        vars = VariableList()
        vars.append(var_1)              # Adds directly a Variable instance
        vars['var/2'] = metadata_2      # Adds the variable with given name and given metadata

    After that, following equalities are True::

        print( var_1 in vars )
        print( 'var/1' in vars.names() )
        print( 'var/2' in vars.names() )

    Note:
        Adding a Variable instance that has a name that is already in the VariableList instance
        will replace the previous Variable instance instead of adding a new one.
    """

    def names(self) -> List[str]:
        """
        :return: names of variables
        """
        return [var.name for var in self]

    def metadata_keys(self) -> List[str]:
        """
        :return: the metadata keys that are common to all variables in the list
        """
        keys = list(self[0].metadata.keys())
        for var in self:
            keys = [key for key in var.metadata.keys() if key in keys]
        return keys

    def append(self, var: Variable) -> None:
        """
        Append var to the end of the list, unless its name is already used. In that case, var
        will replace the previous Variable instance with the same name.
        """
        if not isinstance(var, Variable):
            raise TypeError("VariableList items should be Variable instances")

        if var.name in self.names():
            self[self.names().index(var.name)] = var
        else:
            super().append(var)

    def update(self, other_var_list: "VariableList", add_variables: bool = False):
        """
        Uses variables in other_var_list to update the current VariableList instance.

        For each Variable instance in other_var_list:
            - if a Variable instance with same name exists, it is replaced by the one
              in other_var_list
            - if not, Variable instance from other_var_list will be added only if
              add_variables==True

        :param other_var_list: source for new Variable data
        :param add_variables: if True, variables can be added instead of just updated
        """

        for var in other_var_list:
            if add_variables or var.name in self.names():
                self.append(var)

    def to_ivc(self) -> om.IndepVarComp:
        """
        :return: an OpenMDAO IndepVarComp instance with all variables from current list
        """
        ivc = om.IndepVarComp()
        for variable in self:
            attributes = variable.metadata.copy()
            value = attributes.pop("value")
            # Some attributes are not compatible with add_output
            for attr in METADATA_TO_IGNORE:
                if attr in attributes:
                    del attributes[attr]
            ivc.add_output(variable.name, value, **attributes)

        return ivc

    def to_dataframe(self) -> pd.DataFrame:
        """
        Creates a DataFrame instance from a VariableList instance.

        Column names are "name" + the keys returned by :meth:`Variable.get_authorized_keys`.
        Values in Series "value" are floats or lists (numpy arrays are converted).

        :return: a pandas DataFrame instance with all variables from current list
        """
        var_dict = {"name": []}
        var_dict.update({metadata_name: [] for metadata_name in self.metadata_keys()})

        # To be able to edit floats and integer
        def _check_shape(value):
            if np.shape(value) == (1,):
                value = float(value[0])
            elif np.shape(value) == ():
                pass
            else:
                value = np.asarray(value).tolist()
            return value

        for variable in self:
            value = _check_shape(variable.value)
            var_dict["name"].append(variable.name)
            for metadata_name in self.metadata_keys():
                if metadata_name == "value":
                    var_dict["value"].append(value)
                else:
                    # TODO: make this more generic
                    if metadata_name in ["value", "initial_value", "lower", "upper"]:
                        metadata = _check_shape(variable.metadata[metadata_name])
                    else:
                        metadata = variable.metadata[metadata_name]
                    var_dict[metadata_name].append(metadata)

        df = pd.DataFrame.from_dict(var_dict)

        return df

    @classmethod
    def from_ivc(cls, ivc: om.IndepVarComp) -> "VariableList":
        """
        Creates a VariableList instance from an OpenMDAO IndepVarComp instance

        :param ivc: an IndepVarComp instance
        :return: a VariableList instance
        """
        variables = VariableList()

        if ivc._static_mode:
            var_rel2meta = ivc._static_var_rel2meta
            var_rel_names = ivc._static_var_rel_names
        else:
            var_rel2meta = ivc._var_rel2meta
            var_rel_names = ivc._var_rel_names

        for name in var_rel_names["output"]:
            metadata = var_rel2meta[name]
            value = metadata.pop("value")
            if np.shape(value) == (1,):
                value = float(value[0])
            elif np.shape(value) == ():
                pass
            else:
                value = np.asarray(value)
            metadata.update({"value": value})
            variables[name] = metadata

        return variables

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame) -> "VariableList":
        """
        Creates a VariableList instance from a pandas DataFrame instance.

        The DataFrame instance is expected to have column names "name" + some keys among the ones given by
        :meth:`Variable.get_authorized_keys`.

        :param df: a DataFrame instance
        :return: a VariableList instance
        """
        column_names = [name for name in df.columns]

        # To be able to edit floats and integer
        def _check_shape(value):
            if np.shape(value) == (1,):
                value = float(value[0])
            elif np.shape(value) == ():
                if type(value) == str:
                    value = float(value)
                else:
                    # Integer
                    pass
            else:
                value = np.asarray(value).tolist()
            return value

        def _get_variable(row):
            var_as_dict = {key: val for key, val in zip(column_names, row)}
            # TODO: make this more generic
            for key, val in var_as_dict.items():
                if key in ["value", "initial_value", "lower", "upper"]:
                    var_as_dict[key] = _check_shape(val)
                else:
                    pass
            return Variable(**var_as_dict)

        return VariableList([_get_variable(row) for row in df[column_names].values])

    @classmethod
    def from_system(cls, system: System,) -> "VariableList":
        """
        Creates a VariableList instance containing variables (inputs and outputs)
        of a an OpenMDAO System.
        The inputs (is_input=True) correspond to the variables of IndepVarComp
        components and all the unconnected variables.

        Warning: setup() must NOT have been called.

        In the case of a group, if variables are promoted, the promoted name
        will be used. Otherwise, the absolute name will be used.

        :param system: OpenMDAO Component instance to inspect
        :return: VariableList instance
        """

        problem = om.Problem(deepcopy(system))
        problem.setup()
        return VariableList.from_problem(problem, use_initial_values=True)

    @classmethod
    def from_problem(
        cls, problem: om.Problem, use_initial_values: bool = False, promoted_only=True,
    ) -> "VariableList":
        """
        Creates a VariableList instance containing
        variables (inputs and outputs) of a an OpenMDAO Problem.
        The inputs (is_input=True) correspond to the variables of IndepVarComp
        components and all the unconnected variables.

        If variables are promoted, the promoted name will be used. Otherwise ( and if
        promoted_only is False), the absolute name will be used.

        :param problem: OpenMDAO Problem instance to inspect
        :param use_initial_values: if True, returned instance will contain values before computation
        :param promoted_only: if True, non-promoted variables will be excluded
        :return: VariableList instance
        """
        variables = VariableList()

        # Setup() is needed
        problem = get_problem_after_setup(problem)
        model = problem.model

        # Determining global inputs

        # from unconnected inputs
        mandatory_unconnected, optional_unconnected = get_unconnected_input_names(problem)
        unconnected_abs_names = mandatory_unconnected + optional_unconnected

        unconnected_inputs = []
        for abs_name in unconnected_abs_names:
            unconnected_inputs.append(model._var_abs2prom["input"][abs_name])

        # from ivc outputs
        ivc_inputs = []
        for subsystem in model._subsystems_allprocs:
            if isinstance(subsystem, om.IndepVarComp):
                input_variables = cls.from_ivc(subsystem)
                for var in input_variables:
                    ivc_inputs.append(var.name)

        global_inputs = unconnected_inputs + ivc_inputs

        prom2abs = {}
        prom2abs.update(model._var_allprocs_prom2abs_list["input"])
        prom2abs.update(model._var_allprocs_prom2abs_list["output"])

        for prom_name, abs_names in prom2abs.items():
            if not promoted_only or "." not in prom_name:
                # Pick the first
                abs_name = abs_names[0]
                metadata = model._var_abs2meta[abs_name]

                # Setting type (IN or OUT)
                if prom_name in global_inputs:
                    metadata.update({"is_input": True})
                else:
                    metadata.update({"is_input": False})

                variable = Variable(name=prom_name, **metadata)
                if not use_initial_values:
                    try:
                        # Maybe useless, but we force units to ensure it is consistent
                        variable.value = problem.get_val(prom_name, units=variable.units)
                    except RuntimeError:
                        # In case problem is incompletely set, problem.get_val() will fail.
                        # In such case, falling back to the method for initial values
                        # should be enough.
                        pass
                variables.append(variable)

        return variables

    @classmethod
    def from_unconnected_inputs(
        cls, problem: om.Problem, with_optional_inputs: bool = False
    ) -> "VariableList":
        """
        This function returns a VariableList instance containing
        all the unconnected inputs of an OpenMDAO Problem.

        If *optional_inputs* is False, only inputs that have numpy.nan as default value (hence
        considered as mandatory) will be in returned instance. Otherwise, all unconnected inputs will
        be in returned instance.

        :param problem: OpenMDAO Problem instance to inspect
        :param with_optional_inputs: If True, returned instance will contain all unconnected inputs.
                                Otherwise, it will contain only mandatory ones.
        :return: VariableList instance
        """
        variables = VariableList()

        # Setup() is needed
        problem = get_problem_after_setup(problem)

        mandatory_unconnected, optional_unconnected = get_unconnected_input_names(problem)
        model = problem.model

        # processed_prom_names will store promoted names that have been already processed, so that
        # it won't be stored twice.
        # By processing mandatory variable first, a promoted variable that would be mandatory somewhere
        # and optional elsewhere will be retained as mandatory (and associated value will be NaN),
        # which is fine.
        # For promoted names that link to several optional variables and no mandatory ones, associated
        # value will be the first encountered one, and this is as good a choice as any other.
        processed_prom_names = []

        def _add_outputs(unconnected_names):
            """ Fills ivc with data associated to each provided var"""
            for abs_name in unconnected_names:
                prom_name = model._var_abs2prom["input"][abs_name]
                if prom_name not in processed_prom_names:
                    processed_prom_names.append(prom_name)
                    metadata = model._var_abs2meta[abs_name]
                    metadata.update({"is_input": True})
                    variables[prom_name] = metadata

        _add_outputs(mandatory_unconnected)
        if with_optional_inputs:
            _add_outputs(optional_unconnected)

        return variables

    def __getitem__(self, key) -> Variable:
        if isinstance(key, str):
            return self[self.names().index(key)]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        if isinstance(key, str):
            if isinstance(value, dict):
                if key in self.names():
                    self[key].metadata = value
                else:
                    self.append(Variable(key, **value))
            else:
                raise TypeError(
                    'VariableList can be set with a "string index" only if value is a '
                    "dict of metadata"
                )
        elif not isinstance(value, Variable):
            raise TypeError("VariableList items should be Variable instances")
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key):
        if isinstance(key, str):
            del self[self.names().index(key)]
        else:
            super().__delitem__(key)

    def __add__(self, other) -> Union[List, "VariableList"]:
        if isinstance(other, VariableList):
            return VariableList(super().__add__(other))
        else:
            return super().__add__(other)



"""
    Estimation of vertical tail mean aerodynamic chords
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import math

import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


# TODO: it would be good to have a function to compute MAC for HT, VT and WING
class ComputeVTMAC(ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Vertical tail mean aerodynamic chord estimation """

    def setup(self):
        self.add_input("data:geometry:vertical_tail:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:tip:chord", val=np.nan, units="m")
        self.add_input("data:geometry:vertical_tail:sweep_25", val=np.nan, units="deg")
        self.add_input("data:geometry:vertical_tail:span", val=np.nan, units="m")

        self.add_output("data:geometry:vertical_tail:MAC:length", units="m")
        self.add_output("data:geometry:vertical_tail:MAC:at25percent:x:local", units="m")
        self.add_output("data:geometry:vertical_tail:MAC:z", units="m")

        self.declare_partials(
            "data:geometry:vertical_tail:MAC:length",
            ["data:geometry:vertical_tail:root:chord", "data:geometry:vertical_tail:tip:chord"],
        )
        self.declare_partials(
            "data:geometry:vertical_tail:MAC:at25percent:x:local", "*", method="fd"
        )
        self.declare_partials(
            "data:geometry:vertical_tail:MAC:z",
            [
                "data:geometry:vertical_tail:root:chord",
                "data:geometry:vertical_tail:tip:chord",
                "data:geometry:vertical_tail:span",
            ],
            method="fd",
        )

    def compute(self, inputs, outputs):
        root_chord = inputs["data:geometry:vertical_tail:root:chord"]
        tip_chord = inputs["data:geometry:vertical_tail:tip:chord"]
        sweep_25_vt = inputs["data:geometry:vertical_tail:sweep_25"]
        b_v = inputs["data:geometry:vertical_tail:span"]

        tmp = root_chord * 0.25 + b_v * math.tan(sweep_25_vt / 180.0 * math.pi) - tip_chord * 0.25

        mac_vt = (
            (root_chord ** 2 + root_chord * tip_chord + tip_chord ** 2)
            / (tip_chord + root_chord)
            * 2.0
            / 3.0
        )
        x0_vt = (tmp * (root_chord + 2 * tip_chord)) / (3 * (root_chord + tip_chord))
        z0_vt = (2 * b_v * (0.5 * root_chord + tip_chord)) / (3 * (root_chord + tip_chord))

        outputs["data:geometry:vertical_tail:MAC:length"] = mac_vt
        outputs["data:geometry:vertical_tail:MAC:at25percent:x:local"] = x0_vt
        outputs["data:geometry:vertical_tail:MAC:z"] = z0_vt



"""
    Estimation of nacelle and pylon geometry
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from math import sqrt

import numpy as np
import openmdao.api as om


class ComputeNacelleAndPylonsGeometry(om.ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Nacelle and pylon geometry estimation """

    def setup(self):

        self.add_input("data:propulsion:MTO_thrust", val=np.nan, units="N")
        self.add_input("data:geometry:propulsion:engine:y_ratio", val=np.nan)
        self.add_input("data:geometry:propulsion:layout", val=np.nan)
        self.add_input("data:geometry:wing:span", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:length", val=np.nan, units="m")
        self.add_input("data:geometry:fuselage:maximum_width", val=np.nan, units="m")

        self.add_output("data:geometry:propulsion:pylon:length", units="m")
        self.add_output("data:geometry:propulsion:fan:length", units="m")
        self.add_output("data:geometry:propulsion:nacelle:length", units="m")
        self.add_output("data:geometry:propulsion:nacelle:diameter", units="m")
        self.add_output("data:geometry:landing_gear:height", units="m")
        self.add_output("data:geometry:propulsion:nacelle:y", units="m")
        self.add_output("data:geometry:propulsion:pylon:wetted_area", units="m**2")
        self.add_output("data:geometry:propulsion:nacelle:wetted_area", units="m**2")
        self.add_output("data:weight:propulsion:engine:CG:x", units="m")

        self.declare_partials(
            "data:geometry:propulsion:nacelle:diameter", "data:propulsion:MTO_thrust", method="fd"
        )
        self.declare_partials(
            "data:geometry:propulsion:nacelle:length", "data:propulsion:MTO_thrust", method="fd"
        )
        self.declare_partials(
            "data:geometry:landing_gear:height", "data:propulsion:MTO_thrust", method="fd"
        )
        self.declare_partials(
            "data:geometry:propulsion:fan:length", "data:propulsion:MTO_thrust", method="fd"
        )
        self.declare_partials(
            "data:geometry:propulsion:pylon:length", "data:propulsion:MTO_thrust", method="fd"
        )
        self.declare_partials(
            "data:geometry:propulsion:nacelle:y",
            [
                "data:propulsion:MTO_thrust",
                "data:geometry:fuselage:maximum_width",
                "data:geometry:propulsion:engine:y_ratio",
                "data:geometry:wing:span",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:weight:propulsion:engine:CG:x",
            [
                "data:geometry:wing:MAC:at25percent:x",
                "data:geometry:wing:MAC:length",
                "data:geometry:wing:MAC:leading_edge:x:local",
                "data:geometry:wing:kink:leading_edge:x:local",
                "data:geometry:wing:root:y",
                "data:geometry:wing:kink:y",
                "data:geometry:wing:root:chord",
                "data:geometry:wing:kink:chord",
                "data:geometry:fuselage:length",
                "data:propulsion:MTO_thrust",
                "data:geometry:fuselage:maximum_width",
                "data:geometry:propulsion:engine:y_ratio",
                "data:geometry:wing:span",
            ],
            method="fd",
        )
        self.declare_partials(
            "data:geometry:propulsion:nacelle:wetted_area",
            "data:propulsion:MTO_thrust",
            method="fd",
        )
        self.declare_partials(
            "data:geometry:propulsion:pylon:wetted_area", "data:propulsion:MTO_thrust", method="fd"
        )

    def compute(self, inputs, outputs):
        thrust_sl = inputs["data:propulsion:MTO_thrust"]
        y_ratio_engine = inputs["data:geometry:propulsion:engine:y_ratio"]
        propulsion_layout = np.round(inputs["data:geometry:propulsion:layout"])
        span = inputs["data:geometry:wing:span"]
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        x0_wing = inputs["data:geometry:wing:MAC:leading_edge:x:local"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        l3_wing = inputs["data:geometry:wing:kink:chord"]
        x3_wing = inputs["data:geometry:wing:kink:leading_edge:x:local"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]
        fus_length = inputs["data:geometry:fuselage:length"]
        b_f = inputs["data:geometry:fuselage:maximum_width"]

        nac_dia = 0.00904 * sqrt(thrust_sl * 0.225) + 0.7  # FIXME: use output of engine module
        lg_height = 1.4 * nac_dia
        # The nominal thrust must be used in lbf
        nac_length = 0.032 * sqrt(thrust_sl * 0.225)  # FIXME: use output of engine module

        outputs["data:geometry:propulsion:nacelle:length"] = nac_length
        outputs["data:geometry:propulsion:nacelle:diameter"] = nac_dia
        outputs["data:geometry:landing_gear:height"] = lg_height

        fan_length = 0.60 * nac_length
        pylon_length = 1.1 * nac_length

        if propulsion_layout == 1:
            y_nacell = y_ratio_engine * span / 2
        elif propulsion_layout == 2:
            y_nacell = b_f / 2.0 + 0.5 * nac_dia
        else:
            raise ValueError("Value of data:geometry:propulsion:layout can only be 1 or 2")

        l_wing_nac = l3_wing + (l2_wing - l3_wing) * (y3_wing - y_nacell) / (y3_wing - y2_wing)

        if propulsion_layout == 1:
            delta_x_nacell = 0.05 * l_wing_nac
            x_nacell_cg = (
                x3_wing * (y_nacell - y2_wing) / (y3_wing - y2_wing)
                - delta_x_nacell
                - 0.2 * nac_length
            )
            x_nacell_cg_absolute = fa_length - 0.25 * l0_wing - (x0_wing - x_nacell_cg)
        elif propulsion_layout == 2:
            x_nacell_cg_absolute = 0.8 * fus_length
        else:
            raise ValueError("Value of data:geometry:propulsion:layout can only be 1 or 2")

        outputs["data:geometry:propulsion:pylon:length"] = pylon_length
        outputs["data:geometry:propulsion:fan:length"] = fan_length
        outputs["data:geometry:propulsion:nacelle:y"] = y_nacell
        outputs["data:weight:propulsion:engine:CG:x"] = x_nacell_cg_absolute

        # Wet surfaces
        wet_area_nac = 0.0004 * thrust_sl * 0.225 + 11  # By engine
        wet_area_pylon = 0.35 * wet_area_nac

        outputs["data:geometry:propulsion:nacelle:wetted_area"] = wet_area_nac
        outputs["data:geometry:propulsion:pylon:wetted_area"] = wet_area_pylon



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import pytest
from fastoad.io import VariableIO

from tests.testing_utilities import run_system
from ..tail_sizing.compute_ht_area import ComputeHTArea
from ..tail_sizing.compute_vt_area import ComputeVTArea

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")


# pylint: disable=redefined-outer-name  # needed for pytest fixtures
@pytest.fixture(scope="module")
def input_xml() -> VariableIO:
    """
    :return: access to the sample xml data
    """
    # TODO: have more consistency in input data (no need for the whole geometry_inputs_full.xml)
    return VariableIO(pth.join(DATA_FOLDER_PATH, "hq_inputs.xml"))


def test_compute_ht_area(input_xml):
    """ Tests computation of the horizontal tail area """

    input_list = [
        "data:geometry:has_T_tail",
        "data:geometry:fuselage:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:area",
        "data:weight:airframe:landing_gear:main:CG:x",
        "data:weight:airframe:landing_gear:front:CG:x",
        "data:weight:aircraft:MTOW",
        "settings:weight:aircraft:CG:range",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()
    problem = run_system(ComputeHTArea(), input_vars)

    ht_lp = problem["data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25"]
    assert ht_lp == pytest.approx(17.68, abs=1e-2)
    wet_area = problem["data:geometry:horizontal_tail:wetted_area"]
    assert wet_area == pytest.approx(70.31, abs=1e-2)
    ht_area = problem["data:geometry:horizontal_tail:area"]
    assert ht_area == pytest.approx(35.15, abs=1e-2)


def test_compute_vt_area(input_xml):
    """ Tests computation of the vertical tail area """

    input_list = [
        "data:TLAR:cruise_mach",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:area",
        "data:geometry:wing:span",
        "data:geometry:vertical_tail:MAC:at25percent:x:from_wingMAC25",
        "data:aerodynamics:vertical_tail:cruise:CL_alpha",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft:CG:aft:MAC_position", 0.364924)
    input_vars.add_output("data:aerodynamics:fuselage:cruise:CnBeta", -0.117901)

    component = ComputeVTArea()

    problem = run_system(component, input_vars)

    cn_beta_vt = problem["data:aerodynamics:vertical_tail:cruise:CnBeta"]
    assert cn_beta_vt == pytest.approx(0.258348, abs=1e-6)
    wet_area = problem["data:geometry:vertical_tail:wetted_area"]
    assert wet_area == pytest.approx(52.34, abs=1e-2)
    vt_area = problem["data:geometry:vertical_tail:area"]
    assert vt_area == pytest.approx(24.92, abs=1e-2)



"""
    Estimation of maximum center of gravity ratio
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


class ComputeMaxCGratio(ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Maximum center of gravity ratio estimation """

    def setup(self):
        self.add_input("data:weight:aircraft:empty:CG:MAC_position", val=np.nan)
        self.add_input("data:weight:aircraft:load_case_1:CG:MAC_position", val=np.nan)
        self.add_input("data:weight:aircraft:load_case_2:CG:MAC_position", val=np.nan)
        self.add_input("data:weight:aircraft:load_case_3:CG:MAC_position", val=np.nan)
        self.add_input("data:weight:aircraft:load_case_4:CG:MAC_position", val=np.nan)
        self.add_input(
            "settings:weight:aircraft:CG:aft:MAC_position:margin",
            val=0.05,
            desc="Added margin for getting most aft CG position, "
            "as ratio of mean aerodynamic chord",
        )

        self.add_output("data:weight:aircraft:CG:aft:MAC_position")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        outputs["data:weight:aircraft:CG:aft:MAC_position"] = inputs[
            "settings:weight:aircraft:CG:aft:MAC_position:margin"
        ] + max(
            inputs["data:weight:aircraft:empty:CG:MAC_position"],
            inputs["data:weight:aircraft:load_case_1:CG:MAC_position"],
            inputs["data:weight:aircraft:load_case_2:CG:MAC_position"],
            inputs["data:weight:aircraft:load_case_3:CG:MAC_position"],
            inputs["data:weight:aircraft:load_case_4:CG:MAC_position"],
        )



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from fastoad.models.aerodynamics.components.cd0_fuselage import Cd0Fuselage
from fastoad.models.aerodynamics.components.cd0_ht import Cd0HorizontalTail
from fastoad.models.aerodynamics.components.cd0_nacelle_pylons import Cd0NacelleAndPylons
from fastoad.models.aerodynamics.components.cd0_total import Cd0Total
from fastoad.models.aerodynamics.components.cd0_vt import Cd0VerticalTail
from fastoad.models.aerodynamics.components.cd0_wing import Cd0Wing
from openmdao.core.group import Group


class CD0(Group):
    def setup(self):
        self.add_subsystem("cd0_wing", Cd0Wing(), promotes=["*"])
        self.add_subsystem("cd0_fuselage", Cd0Fuselage(), promotes=["*"])
        self.add_subsystem("cd0_ht", Cd0HorizontalTail(), promotes=["*"])
        self.add_subsystem("cd0_vt", Cd0VerticalTail(), promotes=["*"])
        self.add_subsystem("cd0_nac_pylons", Cd0NacelleAndPylons(), promotes=["*"])
        self.add_subsystem("cd0_total", Cd0Total(), promotes=["*"])



"""Command Line Interface."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import os.path as pth
import shutil
import textwrap
from argparse import ArgumentParser, RawDescriptionHelpFormatter, ArgumentDefaultsHelpFormatter
from distutils.util import strtobool

import fastoad
from fastoad.cmd import api
from fastoad.cmd.exceptions import FastFileExistsError
from fastoad.utils.resource_management.copy import copy_resource_folder

NOTEBOOK_FOLDER_NAME = "FAST_OAD_notebooks"


# TODO: it has become a bit messy down here... Refactoring needed, maybe
#       with a better organization of code by sub-commands


class Main:
    """
    Class for managing command line and doing associated actions
    """

    def __init__(self):
        class _CustomFormatter(RawDescriptionHelpFormatter, ArgumentDefaultsHelpFormatter):
            pass

        self.parser = ArgumentParser(
            description="FAST-OAD main program", formatter_class=_CustomFormatter
        )
        self.problem = None

    # ACTIONS ======================================================================================
    @staticmethod
    def _generate_conf_file(args):
        """Generates a sample TOML file."""
        try:
            api.generate_configuration_file(args.conf_file, args.force)
        except FastFileExistsError:
            if _query_yes_no(
                'Configuration file "%s" already exists. Do you want to overwrite it?'
                % args.conf_file
            ):
                api.generate_configuration_file(args.conf_file, True)
            else:
                print("No file written.")

    @staticmethod
    def _generate_inputs(args):
        """Generates input file according to command line arguments."""
        schema = "legacy" if args.legacy else "native"
        try:
            api.generate_inputs(args.conf_file, args.source, schema, args.force)
        except FastFileExistsError as exc:
            if _query_yes_no(
                'Input file "%s" already exists. Do you want to overwrite it?' % exc.args[1]
            ):
                api.generate_inputs(args.conf_file, args.source, schema, True)
            else:
                print("No file written.")

    @staticmethod
    def _list_systems(args):
        """Prints list of system identifiers."""
        api.list_systems(args.conf_file)

    @staticmethod
    def _list_variables(args):
        """Prints list of system outputs."""
        api.list_variables(args.conf_file)

    @staticmethod
    def _write_n2(args):
        """Creates N2 html file."""
        try:
            api.write_n2(args.conf_file, args.n2_file, args.force)
        except FastFileExistsError:
            if _query_yes_no(
                'N2 file "%s" already exists. Do you want to overwrite it?' % args.n2_file
            ):
                api.write_n2(args.conf_file, args.n2_file, True)
            else:
                print("No file written.")

    @staticmethod
    def _write_xdsm(args):
        """Creates XDSM html file."""

        def _write(overwrite):
            api.write_xdsm(
                args.conf_file,
                args.xdsm_file,
                overwrite=overwrite,
                wop_server_url=args.wop_server,
                depth=args.depth,
            )

        try:
            _write(args.force)
        except FastFileExistsError:
            if _query_yes_no(
                'XDSM file "%s" already exists. Do you want to overwrite it?' % args.xdsm_file
            ):
                _write(True)
            else:
                print("No file written.")

    @staticmethod
    def _evaluate(args):
        """Runs model according to provided problem file."""
        try:
            api.evaluate_problem(args.conf_file, args.force)
        except FastFileExistsError as exc:
            if _query_yes_no(
                'Output file "%s" already exists. Do you want to overwrite it?' % exc.args[1]
            ):
                api.evaluate_problem(args.conf_file, True)
            else:
                print("Computation not run.")

    @staticmethod
    def _optimize(args):
        """Runs driver according to provided problem file."""
        try:
            api.optimize_problem(args.conf_file, args.force)
        except FastFileExistsError as exc:
            if _query_yes_no(
                'Output file "%s" already exists. Do you want to overwrite it?' % exc.args[1]
            ):
                api.optimize_problem(args.conf_file, True)
            else:
                print("Computation not run.")

    @staticmethod
    def _notebooks(args):
        """Copies the notebook folder in user-selected destination."""
        # Create and copy folder
        target_path = pth.abspath(pth.join(args.path, NOTEBOOK_FOLDER_NAME))
        if pth.exists(target_path):
            shutil.rmtree(target_path)
        os.makedirs(target_path)
        copy_resource_folder("fastoad.notebooks.tutorial", target_path)

        # Give info for running Jupyter
        print("")
        print("Notebooks have been created in %s" % target_path)
        print("You may now run Jupyter with:")
        print('   jupyter notebook "%s"' % target_path)

    # UTILITIES ====================================================================================

    # PARSER CONFIGURATION =========================================================================
    @staticmethod
    def _add_conf_file_argument(parser: ArgumentParser, required=True):
        kwargs = {"type": str, "help": "the configuration file for setting the problem"}
        if not required:
            kwargs["nargs"] = "?"
        parser.add_argument("conf_file", **kwargs)

    @staticmethod
    def _add_overwrite_argument(parser: ArgumentParser):
        parser.add_argument(
            "-f", "--force", action="store_true", help="do not ask before overwriting files"
        )

    # ENTRY POINT ==================================================================================
    def run(self):
        """Main function."""
        subparsers = self.parser.add_subparsers(title="sub-commands")

        # option for version -----------------------------------------------------------------------
        self.parser.add_argument(
            "-v",
            "--version",
            action="version",
            version="FAST-OAD " + fastoad.__version__,
            help="shows version number",
        )

        # sub-command for generating sample configuration file -------------------------------------
        parser_gen_conf = subparsers.add_parser(
            "gen_conf",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="generates a sample configuration file",
            description="generates the configuration file with sample data",
        )
        parser_gen_conf.help = parser_gen_conf.description
        parser_gen_conf.add_argument(
            "conf_file", type=str, help="the name of configuration file " "to be written"
        )
        self._add_overwrite_argument(parser_gen_conf)
        parser_gen_conf.set_defaults(func=self._generate_conf_file)

        # sub-command for generating input file ----------------------------------------------------
        class _CustomFormatter(RawDescriptionHelpFormatter, ArgumentDefaultsHelpFormatter):
            pass

        parser_gen_inputs = subparsers.add_parser(
            "gen_inputs",
            formatter_class=_CustomFormatter,
            help="generates the input file",
            description="generates the input file (specified in the configuration file)"
            " with needed variables",
        )
        self._add_conf_file_argument(parser_gen_inputs)
        self._add_overwrite_argument(parser_gen_inputs)
        parser_gen_inputs.add_argument(
            "source",
            nargs="?",
            help="if provided, generated input file will be fed with values from provided XML file",
        )
        parser_gen_inputs.add_argument(
            "--legacy",
            action="store_true",
            help="to be used if the source XML file is in legacy format",
        )
        parser_gen_inputs.set_defaults(func=self._generate_inputs)
        parser_gen_inputs.epilog = textwrap.dedent(
            """\
            Examples:
            ---------
            # For the problem defined in conf_file.toml, generates the input file with default 
            # values (when default values are defined):
                %(prog)s gen_inputs conf_file.toml
            
            # Same as above, except that values are taken from some_file.xml when possible:
                %(prog)s gen_inputs conf_file.toml some_file.xml

            # Same as above, some_file.xml is in the legacy FAST schema
                %(prog)s gen_inputs conf_file.toml some_file.xml --legacy
            """
        )

        # sub-command for listing registered systems -----------------------------------------------
        parser_list_systems = subparsers.add_parser(
            "list_systems",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Provides the identifiers of available systems",
            description="Provides the identifiers of available systems",
        )
        self._add_conf_file_argument(parser_list_systems, required=False)
        parser_list_systems.set_defaults(func=self._list_systems)

        # sub-command for listing possible outputs -------------------------------------------------
        parser_list_variables = subparsers.add_parser(
            "list_variables",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Lists the variables of the problem",
            description="Lists the variables of the problem",
        )
        self._add_conf_file_argument(parser_list_variables)
        parser_list_variables.set_defaults(func=self._list_variables)

        # sub-command for writing N2 diagram -------------------------------------------------------
        parser_n2 = subparsers.add_parser(
            "n2",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Writes the N2 diagram of the problem",
            description="Writes an HTML file that shows the N2 diagram of the problem",
        )
        self._add_conf_file_argument(parser_n2)
        self._add_overwrite_argument(parser_n2)
        parser_n2.add_argument(
            "n2_file", nargs="?", default="n2.html", help="path of file to be written"
        )
        parser_n2.set_defaults(func=self._write_n2)

        # sub-command for writing XDSM diagram -------------------------------------------------------
        parser_xdsm = subparsers.add_parser(
            "xdsm",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Writes the XDSM diagram of the problem",
            description="Writes an HTML file that shows the XDSM diagram of the problem",
        )
        self._add_conf_file_argument(parser_xdsm)
        self._add_overwrite_argument(parser_xdsm)
        parser_xdsm.add_argument(
            "xdsm_file", nargs="?", default="xdsm.html", help="path of file to be written"
        )
        parser_xdsm.add_argument(
            "--server",
            nargs="?",
            default=None,
            dest="wop_server",
            help="WhatsOpt server URL. Needed only at first call",
        )
        parser_xdsm.add_argument(
            "--depth", nargs="?", default=2, help="Depth of analysis", type=int,
        )
        parser_xdsm.set_defaults(func=self._write_xdsm)

        # sub-command for running the model --------------------------------------------------------
        parser_run_model = subparsers.add_parser(
            "eval",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Runs the analysis",
            description="Runs the analysis",
        )
        self._add_conf_file_argument(parser_run_model)
        self._add_overwrite_argument(parser_run_model)
        parser_run_model.set_defaults(func=self._evaluate)

        # sub-command for running the driver -------------------------------------------------------
        parser_run_driver = subparsers.add_parser(
            "optim",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Runs the optimization",
            description="Runs the optimization",
        )
        self._add_conf_file_argument(parser_run_driver)
        self._add_overwrite_argument(parser_run_driver)
        parser_run_driver.set_defaults(func=self._optimize)

        # sub-command for running Jupyter notebooks ------------------------------------------------
        parser_notebooks = subparsers.add_parser(
            "notebooks",
            formatter_class=ArgumentDefaultsHelpFormatter,
            help="Create ready-to-use notebooks",
            description="Creates a %s/ folder with pre-configured Jupyter notebooks. "
            "Please note that an existing FAST_OAD_notebooks/ will be erased"
            % NOTEBOOK_FOLDER_NAME,
        )

        parser_notebooks.add_argument(
            "path",
            default=".",
            nargs="?",
            help="The path where the %s/ folder will be added" % NOTEBOOK_FOLDER_NAME,
        )

        parser_notebooks.set_defaults(func=self._notebooks)

        # Parse ------------------------------------------------------------------------------------
        args = self.parser.parse_args()
        try:
            args.func(args)
        except AttributeError:
            self.parser.print_help()


def _query_yes_no(question):
    """
    Ask a yes/no question via input() and return its answer as boolean.

    Keeps asking while answer is not similar to "yes" or "no"
    The returned value is True for "yes" or False for "no".
    """
    answer = None
    while answer is None:
        raw_answer = input(question + "\n")
        try:
            answer = strtobool(raw_answer)
        except ValueError:
            pass

    return answer == 1


def main():
    log_format = "%(levelname)-8s: %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)
    Main().run()


if __name__ == "__main__":
    main()



"""
Tests custom XML serializer for OpenMDAO variables
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from shutil import rmtree

import pytest
from fastoad.openmdao.variables import VariableList
from pytest import approx

from .. import VariableXmlBaseFormatter
from ..translator import VarXpathTranslator
from ...variable_io import VariableIO

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(
    pth.dirname(__file__), "results", pth.splitext(pth.basename(__file__))[0]
)


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def _check_basic2_vars(outputs: VariableList):
    """ Checks that provided IndepVarComp instance matches content of data/custom.xml file """

    assert len(outputs) == 5

    # Using pytest.approx for numerical reason, but also because it works even if sequence types
    # are different (lists, tuples, numpy arrays)
    assert outputs["geometry:total_surface"].value == approx([780.3])
    assert outputs["geometry:total_surface"].units == "m**2"

    assert outputs["geometry:wing:span"].value == approx([42])
    assert outputs["geometry:wing:span"].units == "m"

    assert outputs["geometry:wing:aspect_ratio"].value == approx([9.8])
    assert outputs["geometry:wing:aspect_ratio"].units is None

    assert outputs["geometry:wing:chord"].value == approx([5.0, 3.5, 2.0])
    assert outputs["geometry:wing:chord"].units == "m"

    assert outputs["geometry:fuselage:length"].value == approx([40.0])
    assert outputs["geometry:fuselage:length"].units == "m"


def test_custom_xml_read_and_write_from_ivc(cleanup):
    """
    Tests the creation of an XML file from an IndepVarComp instance
    """
    result_folder = pth.join(RESULTS_FOLDER_PATH, "custom_xml")

    var_names = [
        "geometry:total_surface",
        "geometry:wing:span",
        "geometry:wing:chord",
        "geometry:wing:aspect_ratio",
        "geometry:fuselage:length",
    ]

    xpaths = ["total_area", "wing/span", "wing/chord", "wing/aspect_ratio", "fuselage_length"]

    # test read ---------------------------------------------------------------

    filename = pth.join(DATA_FOLDER_PATH, "custom.xml")

    translator = VarXpathTranslator(variable_names=var_names, xpaths=xpaths)
    xml_read = VariableIO(filename, formatter=VariableXmlBaseFormatter(translator))
    vars = xml_read.read()
    _check_basic2_vars(vars)

    # test with a non-exhaustive translation table (missing variable name in the translator)
    # we expect that the variable is not included in the ivc
    filename = pth.join(DATA_FOLDER_PATH, "custom_additional_var.xml")
    xml_read = VariableIO(filename, formatter=VariableXmlBaseFormatter(translator))
    vars = xml_read.read()
    _check_basic2_vars(vars)

    # test with setting a translation with an additional var not present in the xml
    filename = pth.join(DATA_FOLDER_PATH, "custom.xml")
    xml_read = VariableIO(
        filename,
        formatter=VariableXmlBaseFormatter(
            VarXpathTranslator(
                variable_names=var_names + ["additional_var"], xpaths=xpaths + ["bad:xpath"]
            )
        ),
    )
    vars = xml_read.read()
    _check_basic2_vars(vars)

    # test write --------------------------------------------------------------
    new_filename = pth.join(result_folder, "custom.xml")
    translator = VarXpathTranslator(variable_names=var_names, xpaths=xpaths)
    xml_write = VariableIO(new_filename, formatter=VariableXmlBaseFormatter(translator))
    xml_write.write(vars)

    # check written data
    assert pth.isfile(new_filename)
    translator.set(var_names, xpaths)
    xml_check = VariableIO(new_filename, formatter=VariableXmlBaseFormatter(translator))
    new_ivc = xml_check.read()
    _check_basic2_vars(new_ivc)


def test_custom_xml_read_and_write_with_translation_table(cleanup):
    """
    Tests the creation of an XML file with a translation table
    """
    result_folder = pth.join(RESULTS_FOLDER_PATH, "custom_xml_with_translation_table")

    # test read ---------------------------------------------------------------
    filename = pth.join(DATA_FOLDER_PATH, "custom.xml")

    # test after setting translation table
    translator = VarXpathTranslator(source=pth.join(DATA_FOLDER_PATH, "custom_translation.txt"))
    xml_read = VariableIO(filename, formatter=VariableXmlBaseFormatter(translator))
    vars = xml_read.read()
    _check_basic2_vars(vars)

    new_filename = pth.join(result_folder, "custom.xml")
    xml_write = VariableIO(new_filename, formatter=VariableXmlBaseFormatter(translator))
    xml_write.write(vars)


def test_custom_xml_read_and_write_with_only_or_ignore(cleanup):
    """
    Tests the creation of an XML file with only and ignore options
    """

    var_names = [
        "geometry:total_surface",
        "geometry:wing:span",
        "geometry:wing:aspect_ratio",
        "geometry:fuselage:length",
    ]

    xpaths = ["total_area", "wing/span", "wing/aspect_ratio", "fuselage_length"]

    # test read ---------------------------------------------------------------
    filename = pth.join(DATA_FOLDER_PATH, "custom.xml")

    translator = VarXpathTranslator(variable_names=var_names, xpaths=xpaths)
    xml_read = VariableIO(filename, formatter=VariableXmlBaseFormatter(translator))

    # test with "only"
    outputs = xml_read.read(only=["geometry:wing:span"])
    assert len(outputs) == 1
    assert outputs[0].name == "geometry:wing:span"
    assert outputs[0].value == approx([42])
    assert outputs[0].units == "m"

    # test with "ignore"
    outputs = xml_read.read(
        ignore=["geometry:total_surface", "geometry:wing:aspect_ratio", "geometry:fuselage:length"]
    )
    assert len(outputs) == 1
    assert outputs[0].name == "geometry:wing:span"
    assert outputs[0].value == approx([42])
    assert outputs[0].units == "m"

    # test with patterns in "only"
    outputs = xml_read.read(only=["*:wing:*"])
    assert len(outputs) == 2
    assert outputs[0].name == "geometry:wing:span"
    assert outputs[0].value == approx([42])
    assert outputs[0].units == "m"
    assert outputs[1].name == "geometry:wing:aspect_ratio"
    assert outputs[1].value == approx([9.8])
    assert outputs[1].units is None

    # test with patterns in "ignore"
    outputs = xml_read.read(ignore=["geometry:*u*", "geometry:wing:aspect_ratio"])
    assert len(outputs) == 1
    assert outputs[0].name == "geometry:wing:span"
    assert outputs[0].value == approx([42])
    assert outputs[0].units == "m"



"""Base module for propulsion models."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Union, Sequence, Optional, Tuple

import numpy as np
import openmdao.api as om
from fastoad.constants import FlightPhase
from openmdao.core.component import Component


class IPropulsion(ABC):
    """
    Interface that should be implemented by propulsion models.
    """

    @abstractmethod
    def compute_flight_points(
        self,
        mach: Union[float, Sequence],
        altitude: Union[float, Sequence],
        phase: Union[FlightPhase, Sequence],
        use_thrust_rate: Optional[Union[bool, Sequence]] = None,
        thrust_rate: Optional[Union[float, Sequence]] = None,
        thrust: Optional[Union[float, Sequence]] = None,
    ) -> Tuple[Union[float, Sequence], Union[float, Sequence], Union[float, Sequence]]:
        """
        Computes Specific Fuel Consumption according to provided conditions.

        Each input can be a float, a list or an array.
        Inputs that are not floats must all have the same shape (numpy speaking).

        About use_thrust_rate, thrust_rate and thrust
        ---------------------------------------------

            *use_thrust_rate* tells if a flight point should be computed using *thrust_rate*
            or *thrust* as input.

            - if *use_thrust_rate* is None, the considered input will be the provided one
            between *thrust_rate* and *thrust* (if both are provided, *thrust_rate* will be used)

            - if *use_thrust_rate* is True or False (i.e., not a sequence), the considered input
            will be taken accordingly, and should of course not be None.

            - if *use_thrust_rate* is a sequence or array, *thrust_rate* and *thrust* should be
            provided and have the same shape as *use_thrust_rate*. The method will consider for
            each element which input will be used according to *use_thrust_rate*


        :param mach: Mach number
        :param altitude: (unit=m) altitude w.r.t. to sea level
        :param phase: flight phase
        :param use_thrust_rate: tells if thrust_rate or thrust should be used (works element-wise)
        :param thrust_rate: thrust rate (unit=none)
        :param thrust: required thrust (unit=N)
        :return: SFC (in kg/s/N), thrust rate, thrust (in N)
        """


class EngineSet(IPropulsion):
    def __init__(self, engine: IPropulsion, engine_count):
        """
        Class for modelling an assembly of identical engines.

        Thrust is supposed equally distributed among them.

        :param engine: the engine model
        :param engine_count:
        """
        self.engine = engine
        self.engine_count = engine_count

    def compute_flight_points(
        self,
        mach: Union[float, Sequence],
        altitude: Union[float, Sequence],
        phase: Union[FlightPhase, Sequence],
        use_thrust_rate: Optional[Union[bool, Sequence]] = None,
        thrust_rate: Optional[Union[float, Sequence]] = None,
        thrust: Optional[Union[float, Sequence]] = None,
    ) -> Tuple[Union[float, Sequence], Union[float, Sequence], Union[float, Sequence]]:

        if thrust is not None:
            thrust = thrust / self.engine_count

        sfc, thrust_rate, thrust = self.engine.compute_flight_points(
            mach, altitude, phase, use_thrust_rate, thrust_rate, thrust
        )
        return sfc, thrust_rate, thrust * self.engine_count


class IOMPropulsionWrapper:
    """
    Interface for wrapping a :class:`IPropulsion` subclass in OpenMDAO.

    The implementation class defines the needed input variables for instantiating the
    :class:`IPropulsion` subclass in :meth:`setup` and use them for instantiation in
    :meth:`get_model`

    see :class:`~fastoad.models.propulsion.fuel_propulsion.rubber_engine.OMRubberEngineWrapper` for
    an example of implementation.
    """

    @abstractmethod
    def setup(self, component: Component):
        """
        Defines the needed OpenMDAO inputs for propulsion instantiation as done in :meth:`get_model`

        Use `add_inputs` and `declare_partials` methods of the provided `component`

        :param component:
        """

    @staticmethod
    @abstractmethod
    def get_model(inputs) -> IPropulsion:
        """
        This method defines the used :class:`IPropulsion` subclass instance.

        :param inputs: OpenMDAO input vector where the parameters that define the
                       propulsion model are
        :return: the propulsion model instance
        """


class BaseOMPropulsionComponent(om.ExplicitComponent, ABC):
    """
    Base class for OpenMDAO wrapping of subclasses of :class:`IEngineForOpenMDAO`.

    Classes that implements this interface should add their own inputs in setup()
    and implement :meth:`get_wrapper`.
    """

    def initialize(self):
        self.options.declare("flight_point_count", 1, types=(int, tuple))

    def setup(self):
        shape = self.options["flight_point_count"]
        self.add_input("data:propulsion:mach", np.nan, shape=shape)
        self.add_input("data:propulsion:altitude", np.nan, shape=shape, units="m")
        self.add_input("data:propulsion:phase", np.nan, shape=shape)
        self.add_input("data:propulsion:use_thrust_rate", np.nan, shape=shape)
        self.add_input("data:propulsion:required_thrust_rate", np.nan, shape=shape)
        self.add_input("data:propulsion:required_thrust", np.nan, shape=shape, units="N")

        self.add_output("data:propulsion:SFC", shape=shape, units="kg/s/N", ref=1e-4)
        self.add_output("data:propulsion:thrust_rate", shape=shape, lower=0.0, upper=1.0)
        self.add_output("data:propulsion:thrust", shape=shape, units="N", ref=1e5)

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        wrapper = self.get_wrapper().get_model(inputs)
        sfc, thrust_rate, thrust = wrapper.compute_flight_points(
            inputs["data:propulsion:mach"],
            inputs["data:propulsion:altitude"],
            inputs["data:propulsion:phase"],
            inputs["data:propulsion:use_thrust_rate"],
            inputs["data:propulsion:required_thrust_rate"],
            inputs["data:propulsion:required_thrust"],
        )
        outputs["data:propulsion:SFC"] = sfc
        outputs["data:propulsion:thrust_rate"] = thrust_rate
        outputs["data:propulsion:thrust"] = thrust

    @staticmethod
    @abstractmethod
    def get_wrapper() -> IOMPropulsionWrapper:
        """
        This method defines the used :class:`IOMPropulsionWrapper` instance.

        :return: an instance of OpenMDAO wrapper for propulsion model
        """



"""
Test module for XFOIL component
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=redefined-outer-name  # needed for fixtures

import os.path as pth
import shutil
from platform import system

import pytest
from openmdao.core.indepvarcomp import IndepVarComp

from tests.testing_utilities import run_system
from tests.xfoil_exe.get_xfoil import get_xfoil_path
from ..xfoil_polar import XfoilPolar, DEFAULT_2D_CL_MAX

XFOIL_RESULTS = pth.join(pth.dirname(__file__), "results")

xfoil_path = None if system() == "Windows" else get_xfoil_path()


@pytest.mark.skipif(
    system() != "Windows" and xfoil_path is None, reason="No XFOIL executable available"
)
def test_compute():
    """ Tests a simple XFOIL run"""

    if pth.exists(XFOIL_RESULTS):
        shutil.rmtree(XFOIL_RESULTS)

    ivc = IndepVarComp()
    ivc.add_output("xfoil:reynolds", 18000000)
    ivc.add_output("xfoil:mach", 0.20)
    ivc.add_output("data:geometry:wing:thickness_ratio", 0.1284)

    xfoil_comp = XfoilPolar(
        alpha_start=15.0, alpha_end=25.0, iter_limit=20, xfoil_exe_path=xfoil_path
    )
    problem = run_system(xfoil_comp, ivc)
    assert problem["xfoil:CL_max_2D"] == pytest.approx(1.94, 1e-2)
    assert not pth.exists(XFOIL_RESULTS)

    xfoil_comp = XfoilPolar(
        alpha_start=12.0, alpha_end=20.0, iter_limit=20, xfoil_exe_path=xfoil_path
    )  # will stop before real max CL
    problem = run_system(xfoil_comp, ivc)
    assert problem["xfoil:CL_max_2D"] == pytest.approx(1.92, 1e-2)
    assert not pth.exists(XFOIL_RESULTS)

    xfoil_comp = XfoilPolar(
        alpha_start=50.0, alpha_end=55.0, iter_limit=2, xfoil_exe_path=xfoil_path
    )  # will not converge
    problem = run_system(xfoil_comp, ivc)
    assert problem["xfoil:CL_max_2D"] == pytest.approx(DEFAULT_2D_CL_MAX, 1e-2)
    assert not pth.exists(XFOIL_RESULTS)

    xfoil_comp = XfoilPolar(
        iter_limit=20, result_folder_path=XFOIL_RESULTS, xfoil_exe_path=xfoil_path
    )
    problem = run_system(xfoil_comp, ivc)
    assert problem["xfoil:CL_max_2D"] == pytest.approx(1.94, 1e-2)
    assert pth.exists(XFOIL_RESULTS)
    assert pth.exists(pth.join(XFOIL_RESULTS, "polar_result.txt"))


@pytest.mark.skipif(
    system() != "Windows" and xfoil_path is None, reason="No XFOIL executable available"
)
def test_compute_with_provided_path():
    """ Test that option "use_exe_path" works """
    ivc = IndepVarComp()
    ivc.add_output("xfoil:reynolds", 18000000)
    ivc.add_output("xfoil:mach", 0.20)
    ivc.add_output("data:geometry:wing:thickness_ratio", 0.1284)

    xfoil_comp = XfoilPolar(alpha_start=18.0, alpha_end=21.0, iter_limit=20)
    xfoil_comp.options["xfoil_exe_path"] = "Dummy"  # bad name
    with pytest.raises(ValueError):
        problem = run_system(xfoil_comp, ivc)

    xfoil_comp.options["xfoil_exe_path"] = (
        xfoil_path
        if xfoil_path
        else pth.join(pth.dirname(__file__), pth.pardir, "xfoil699", "xfoil.exe")
    )
    problem = run_system(xfoil_comp, ivc)
    assert problem["xfoil:CL_max_2D"] == pytest.approx(1.94, 1e-2)



"""
Where initialization operations can be done
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.register import register_openmdao_systems
from pelix.constants import BundleActivator


@BundleActivator
class FASTOADActivator:
    """
    Activator class where internal modules are loaded

    Having the loading operation in start() method ensures it is done only once
    as long as this iPOPO bundle is active.
    """

    def start(self, context):
        """
        Loads inner packages of FAST
        """
        register_openmdao_systems()



"""
Test module for openmdao_system_registry.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os.path as pth

import pytest
from openmdao.api import Problem, ScipyOptimizeDriver

from .sellar_example.disc2 import Disc2
from .sellar_example.sellar import Sellar, ISellarFactory
from .. import BundleLoader
from .. import OpenMDAOSystemRegistry as Registry
from ..constants import SERVICE_OPENMDAO_SYSTEM, ModelDomain
from ..exceptions import FastUnknownOMSystemIdentifierError, FastBadSystemOptionError

_LOGGER = logging.getLogger(__name__)
"""Logger for this module"""

logging.basicConfig(level=logging.DEBUG)


# pylint: disable=redefined-outer-name  # pytest fixture
# pylint: disable=unused-argument  # pytest fixture


@pytest.fixture(scope="module")
def load():
    """ Loads components """
    Registry.explore_folder(pth.join(pth.dirname(__file__), "sellar_example"))


def test_components_alone(load):
    """
    Simple test of existence of "openmdao.component" service factories
    """

    services = BundleLoader().get_factory_names(SERVICE_OPENMDAO_SYSTEM)
    assert services


def test_get_system(load):
    """
    Tests the retrieval of a System according to identifier
    """

    # Get component 1 #########################################################
    # Tests also the definition of model domain
    disc1_component = Registry.get_system("sellar.disc1")
    assert Registry.get_system_domain("sellar.disc1").value == ModelDomain.OTHER.value
    assert Registry.get_system_domain(disc1_component).value == ModelDomain.OTHER.value
    assert Registry.get_system_description(disc1_component) == "some text"
    assert disc1_component is not None
    disc1_component.setup()
    outputs = {}
    disc1_component.compute({"z": [10.0, 10.0], "x": 10.0, "y2": 10.0}, outputs)
    assert outputs["y1"] == 118.0

    # Get component 2 #########################################################
    # Tests also the transmission of options
    with pytest.raises(FastBadSystemOptionError):
        disc2_component = Registry.get_system("sellar.disc2", options={"not_declared": -1})

    disc2_component = Registry.get_system("sellar.disc2", options={"answer": -1})
    assert Registry.get_system_domain("sellar.disc2").value == ModelDomain.GEOMETRY.value
    assert Registry.get_system_domain(disc2_component).value == ModelDomain.GEOMETRY.value
    assert Registry.get_system_description(disc2_component) == Disc2.__doc__
    assert (
        disc2_component.options["answer"] == 42
    )  # still the initial value as setup() has not been run
    disc2_component.setup()
    assert disc2_component.options["answer"] == -1
    outputs = {}
    disc2_component.compute({"z": [10.0, 10.0], "y1": 4.0}, outputs)
    assert outputs["y2"] == 22.0

    # Get component 2 bis #####################################################
    # Tests the transmission of options at registration
    with pytest.raises(FastBadSystemOptionError):
        functions_component = Registry.get_system("sellar.functions", options={"not_declared": -1})

    functions_component = Registry.get_system("sellar.functions")
    assert Registry.get_system_domain("sellar.functions").value == ModelDomain.UNSPECIFIED.value
    assert Registry.get_system_domain(functions_component).value == ModelDomain.UNSPECIFIED.value
    assert (
        functions_component.options["best_doctor"] == 10
    )  # still the initial value as setup() has not been run
    functions_component.setup()
    assert functions_component.options["best_doctor"] == 11

    # Get unknown component ###################################################
    with pytest.raises(FastUnknownOMSystemIdentifierError):
        Registry.get_system("unknown.identifier")


def test_sellar(load):
    """
    Demonstrates usage of OpenMDAOSystemRegistry in a simple Sellar problem
    """

    def sellar_setup(sellar_instance: Sellar):
        """
        Sets up the Sellar problem, given a Sellar Group() instance.

        Pure OpenMDAO scripting.
        """
        problem = Problem()
        problem.model = sellar_instance

        problem.driver = ScipyOptimizeDriver()

        problem.driver.options["optimizer"] = "SLSQP"
        problem.driver.options["tol"] = 1.0e-08
        # pb.driver.options['maxiter'] = 100
        problem.driver.options["disp"] = True

        problem.model.add_design_var("x", lower=0, upper=10)
        problem.model.add_design_var("z", lower=0, upper=10)

        problem.model.add_objective("f")

        problem.model.add_constraint("g1", upper=0.0)
        problem.model.add_constraint("g2", upper=0.0)

        problem.setup()

        return problem

    class SellarComponentProviderByFast(ISellarFactory):
        """
        Provides Sellar components using OpenMDAOSystemRegistry
        """

        @staticmethod
        def create_disc1():
            return Registry.get_system("sellar.disc1")

        @staticmethod
        def create_disc2():
            return Registry.get_system("sellar.disc2")

        @staticmethod
        def create_functions():
            return Registry.get_system("sellar.functions")

    classical_problem = sellar_setup(Sellar())  # Reference
    fastoad_problem = sellar_setup(
        Sellar(SellarComponentProviderByFast)
    )  # Using OpenMDAOSystemRegistry

    classical_problem.run_driver()
    assert classical_problem["f"] != fastoad_problem["f"]  # fastoad_problem has not run yet

    fastoad_problem.run_driver()
    assert classical_problem["f"] == fastoad_problem["f"]  # both problems have run



"""
Tests for atmosphere.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numbers import Number

import numpy as np
import pytest
from scipy.constants import foot

from ..atmosphere import Atmosphere


def test_atmosphere():
    """ Tests properties of Atmosphere class """
    # Altitudes in meters Values at disa=0 from "Advanced Aircraft Design (
    # Egbert TORENBEEK, Oxford, UK: John Wiley & Sons Ltd, 2013) Appendix B,
    # p.397-398" Values at disa=10 from
    # https://www.digitaldutch.com/atmoscalc/, with a 0.98749 factor on
    # viscosity because at sea level and disa=0, the calculator gives
    # 1.81206e-5 for dynamic viscosity, though ISA assumption is 1.7894e-5
    expectations = np.array(
        [
            (0, 0, 288.15, 1.225, 101325, 1.460e-05, 340.29),
            (500, 0, 284.90, 1.1673, 95461, 1.519e-05, 338.37),
            (1000, 0, 281.65, 1.1117, 89874, 1.581e-05, 336.43),
            (1500, 0, 278.40, 1.0581, 84556, 1.646e-05, 334.49),
            (2000, 0, 275.15, 1.0065, 79495, 1.714e-05, 332.53),
            (2500, 0, 271.90, 0.9569, 74682, 1.787e-05, 330.56),
            (3000, 0, 268.65, 0.9091, 70108, 1.863e-05, 328.58),
            (3500, 0, 265.40, 0.8632, 65764, 1.943e-05, 326.58),
            (4000, 0, 262.15, 0.8191, 61640, 2.028e-05, 324.58),
            (4500, 0, 258.90, 0.7768, 57728, 2.117e-05, 322.56),
            (5000, 0, 255.65, 0.7361, 54020, 2.211e-05, 320.53),
            (5500, 0, 252.40, 0.6971, 50506, 2.311e-05, 318.48),
            (6000, 0, 249.15, 0.6597, 47181, 2.417e-05, 316.43),
            (6500, 0, 245.90, 0.6238, 44034, 2.529e-05, 314.36),
            (7000, 0, 242.65, 0.5895, 41060, 2.648e-05, 312.27),
            (7500, 0, 239.40, 0.5566, 38251, 2.773e-05, 310.17),
            (8000, 0, 236.15, 0.5252, 35599, 2.906e-05, 308.06),
            (8500, 0, 232.90, 0.4951, 33099, 3.048e-05, 305.93),
            (9000, 0, 229.65, 0.4663, 30742, 3.199e-05, 303.79),
            (9500, 0, 226.40, 0.4389, 28523, 3.359e-05, 301.63),
            (10000, 0, 223.15, 0.4127, 26436, 3.530e-05, 299.46),
            (10500, 0, 219.90, 0.3877, 24474, 3.712e-05, 297.27),
            (11000, 0, 216.65, 0.3639, 22632, 3.905e-05, 295.07),
            (12000, 0, 216.65, 0.3108, 19330, 4.573e-05, 295.07),
            (13000, 0, 216.65, 0.2655, 16510, 5.353e-05, 295.07),
            (14000, 0, 216.65, 0.2268, 14101, 6.266e-05, 295.07),
            (15000, 0, 216.65, 0.1937, 12044, 7.337e-05, 295.07),
            (16000, 0, 216.65, 0.1654, 10287, 8.592e-05, 295.07),
            (17000, 0, 216.65, 0.1413, 8786, 1.006e-04, 295.07),
            (18000, 0, 216.65, 0.1207, 7505, 1.177e-04, 295.07),
            (19000, 0, 216.65, 0.1031, 6410, 1.378e-04, 295.07),
            (20000, 0, 216.65, 0.088, 5475, 1.615e-04, 295.07),
            (0, 10, 298.15, 1.1839, 101325, 1.5527e-5, 346.15),
            (1000, 10, 291.65, 1.0735, 89875, 1.6829e-5, 342.36),
            (3000, 10, 278.65, 0.87650, 70108, 1.9877e-5, 334.64),
            (10000, 10, 233.15, 0.39500, 26436, 3.8106e-05, 306.10),
            (14000, 10, 226.65, 0.2167, 14102, 6.7808e-05, 301.80),
        ],
        dtype=[
            ("alt", "f8"),
            ("dT", "f4"),
            ("T", "f4"),
            ("rho", "f4"),
            ("P", "f4"),
            ("visc", "f4"),
            ("SoS", "f4"),
        ],
    )

    for values in expectations:
        # Checking with altitude provided as scalar
        alt = values["alt"] / foot
        assert isinstance(alt, Number)
        atm = Atmosphere(alt, values["dT"])
        assert values["T"] == pytest.approx(atm.temperature, rel=1e-4)
        assert values["rho"] == pytest.approx(atm.density, rel=1e-3)
        assert values["P"] == pytest.approx(atm.pressure, rel=1e-4)
        assert values["visc"] == pytest.approx(atm.kinematic_viscosity, rel=1e-2)
        assert values["SoS"] == pytest.approx(atm.speed_of_sound, rel=1e-3)

        # Checking with altitude provided as one-element list
        alt = [values["alt"] / foot]
        assert isinstance(alt, list)
        atm = Atmosphere(alt, values["dT"])
        assert values["T"] == pytest.approx(atm.temperature, rel=1e-4)
        assert values["rho"] == pytest.approx(atm.density, rel=1e-3)
        assert values["P"] == pytest.approx(atm.pressure, rel=1e-4)
        assert values["visc"] == pytest.approx(atm.kinematic_viscosity, rel=1e-2)
        assert values["SoS"] == pytest.approx(atm.speed_of_sound, rel=1e-3)

    for delta_t in [0, 10]:
        idx = expectations["dT"] == delta_t

        # Checking with altitude provided as 1D numpy array
        alt = expectations["alt"][idx] / foot
        assert isinstance(alt, np.ndarray)
        assert len(alt.shape) == 1
        atm = Atmosphere(alt, delta_t)
        assert expectations["T"][idx] == pytest.approx(atm.temperature, rel=1e-4)
        assert expectations["rho"][idx] == pytest.approx(atm.density, rel=1e-3)
        assert expectations["P"][idx] == pytest.approx(atm.pressure, rel=1e-4)
        assert expectations["visc"][idx] == pytest.approx(atm.kinematic_viscosity, rel=1e-2)
        assert expectations["SoS"][idx] == pytest.approx(atm.speed_of_sound, rel=1e-3)
        # Additional check for get_altitude in meters
        assert expectations["alt"][idx] == pytest.approx(
            atm.get_altitude(altitude_in_feet=False), rel=1e-3
        )

        # Checking with altitude provided as a list and in meters
        alt = expectations["alt"][idx].tolist()
        assert isinstance(alt, list)
        atm = Atmosphere(alt, delta_t, altitude_in_feet=False)
        assert expectations["T"][idx] == pytest.approx(atm.temperature, rel=1e-4)
        assert expectations["rho"][idx] == pytest.approx(atm.density, rel=1e-3)
        assert expectations["P"][idx] == pytest.approx(atm.pressure, rel=1e-4)
        assert expectations["visc"][idx] == pytest.approx(atm.kinematic_viscosity, rel=1e-2)
        assert expectations["SoS"][idx] == pytest.approx(atm.speed_of_sound, rel=1e-3)
        # Additional check for get_altitude in feet
        assert expectations["alt"][idx] / foot == pytest.approx(atm.get_altitude(), rel=1e-3)


def test_reynolds():
    """ Tests computation of Reynolds number """
    atm = Atmosphere([[0, 35000], [0, 35000]])
    mach = [[0.2, 0.2], [0.8, 0.8]]

    # source:  http://www.aerospaceweb.org/design/scripts/atmosphere/
    expected_reynolds = [[4.6593e6, 1.5738e6], [1.8637e7, 6.2952e6]]

    for result, expected_values in zip(atm.get_unitary_reynolds(mach), expected_reynolds):
        assert result == pytest.approx(expected_values, 5e-3)



"""
Module for management of options and factorizing their definition.
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import openmdao.api as om

CABIN_SIZING_OPTION = "cabin_sizing"
PAYLOAD_FROM_NPAX = "payload_from_npax"


class OpenMdaoOptionDispatcherGroup(om.Group):
    """
    Helper class for transmitting option values to subsystems during self.configure()

    Just create a group by inheriting of this class instead of om.Group. Any option that is
    defined in the group will be transmitted to its immediate subsystems (no recursive
    behaviour), if they have the same option.
    """

    def configure(self):
        """ Update options for all subsystems """
        for key in self.options:
            value = self.options[key]
            for subsystem in self.system_iter(recurse=False):
                if key in subsystem.options:
                    subsystem.options[key] = value



"""
Tests for FAST-OAD optimization viewer
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from shutil import rmtree, copyfile

import pytest
from fastoad.cmd import api
from fastoad.io.configuration.configuration import FASTOADProblemConfigurator

from .. import OptimizationViewer
from ..exceptions import FastMissingFile

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")
RESULTS_FOLDER_PATH = pth.join(pth.dirname(__file__), "results")


@pytest.fixture(scope="module")
def cleanup():
    rmtree(RESULTS_FOLDER_PATH, ignore_errors=True)


def test_optimization_viewer_load(cleanup):
    """
    Basic tests for testing the OptimizationViewer load method.
    """
    filename = pth.join(DATA_FOLDER_PATH, "valid_sellar.toml")

    # The problem has not yet been run
    problem_configuration = FASTOADProblemConfigurator(filename)

    optim_viewer = OptimizationViewer()

    # No input file exists
    with pytest.raises(FastMissingFile):
        optim_viewer.load(problem_configuration)

    api.generate_inputs(filename, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True)

    # Input file exist
    optim_viewer.load(problem_configuration)

    # We run the problem
    api.optimize_problem(filename, overwrite=True)

    # Load the results
    optim_viewer.load(problem_configuration)


def test_optimization_viewer_save(cleanup):
    """
    Basic tests for testing the OptimizationViewer save method.
    """
    filename = pth.join(DATA_FOLDER_PATH, "valid_sellar.toml")
    new_filename = pth.join(RESULTS_FOLDER_PATH, "new_valid_sellar.toml")
    copyfile(filename, new_filename)

    # Loading new file
    problem_configuration = FASTOADProblemConfigurator(filename)

    optim_viewer = OptimizationViewer()

    api.generate_inputs(new_filename, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True)

    # Load new file
    optim_viewer.load(problem_configuration)
    optim_viewer.save()
    optim_viewer.load(problem_configuration)

    # We run the problem
    api.optimize_problem(new_filename, overwrite=True)
    optim_viewer.load(problem_configuration)
    optim_viewer.save()
    optim_viewer.load(problem_configuration)


def test_optimization_viewer_display(cleanup):
    """
    Basic tests for testing the OptimizationViewer load method.
    """
    filename = pth.join(DATA_FOLDER_PATH, "valid_sellar.toml")

    # The problem has not yet been ran
    problem_configuration = FASTOADProblemConfigurator(filename)

    optim_viewer = OptimizationViewer()

    api.generate_inputs(filename, pth.join(DATA_FOLDER_PATH, "inputs.xml"), overwrite=True)

    optim_viewer.load(problem_configuration)
    optim_viewer.display()



"""
Test module for translator.py
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import pytest

from ..exceptions import (
    FastXpathTranslatorInconsistentLists,
    FastXpathTranslatorDuplicates,
    FastXpathTranslatorXPathError,
    FastXpathTranslatorVariableError,
)
from ..translator import VarXpathTranslator


def test_translator_with_set():
    """ Tests VarXpathTranslator using set() for providing translation data"""

    translator = VarXpathTranslator()
    indices = range(10)
    var_list = ["var%i" % i for i in indices]
    xpath_list = ["xpath%i" % i for i in indices]

    # test with lists of different lengths -> error
    var_list2 = ["var0"] + var_list
    with pytest.raises(FastXpathTranslatorInconsistentLists):
        translator.set(var_list2, xpath_list)

    # test with duplicate var names -> error
    duplicate_vars = ["var1", "var3", "var3"]
    other_xpaths = ["xpath42", "xpath404", "xpath0"]
    var_list3 = var_list + duplicate_vars
    xpath_list3 = xpath_list + other_xpaths
    with pytest.raises(FastXpathTranslatorDuplicates) as exc_info:
        translator.set(var_list3, xpath_list3)
    assert exc_info.value.args[1] == set(duplicate_vars)

    # test with duplicate XPaths -> error
    other_vars = ["var42", "var404"]
    duplicate_xpaths = ["xpath5", "xpath2"]
    var_list4 = var_list + other_vars
    xpath_list4 = xpath_list + duplicate_xpaths
    with pytest.raises(FastXpathTranslatorDuplicates) as exc_info:
        translator.set(var_list4, xpath_list4)
    assert exc_info.value.args[1] == set(duplicate_xpaths)

    # Filling correct lists
    translator.set(var_list, xpath_list)

    assert translator.variable_names == var_list
    assert translator.xpaths == xpath_list

    for i in indices:
        assert translator.get_xpath("var%i" % i) == "xpath%i" % i
        assert translator.get_variable_name("xpath%i" % i) == "var%i" % i

    with pytest.raises(FastXpathTranslatorVariableError) as exc_info:
        _ = translator.get_xpath("unknown_var")
    assert exc_info is not None

    with pytest.raises(FastXpathTranslatorXPathError) as exc_info:
        _ = translator.get_variable_name("unknown_path")
    assert exc_info is not None


def test_translator_with_read():
    """ Tests VarXpathTranslator using read() for providing translation data"""

    data_file = pth.join(pth.dirname(__file__), "data", "custom_translation.txt")
    translator = VarXpathTranslator()
    translator.read_translation_table(data_file)

    var_list = [
        "geometry:total_surface",
        "geometry:wing:span",
        "geometry:wing:chord",
        "geometry:wing:aspect_ratio",
        "geometry:fuselage:length",
    ]
    xpath_list = ["total_area", "wing/span", "wing/chord", "wing/aspect_ratio", "fuselage_length"]

    assert translator.variable_names == var_list
    assert translator.xpaths == xpath_list

    for var_name, xpath in zip(var_list, xpath_list):
        assert translator.get_variable_name(xpath) == var_name
        assert translator.get_xpath(var_name) == xpath



"""
    FAST - Copyright (c) 2016 ONERA ISAE
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from fastoad.models.aerodynamics.constants import POLAR_POINT_COUNT
from openmdao.core.explicitcomponent import ExplicitComponent


class CdTrim(ExplicitComponent):
    def initialize(self):
        self.options.declare("low_speed_aero", default=False, types=bool)

    def setup(self):
        self.low_speed_aero = self.options["low_speed_aero"]

        nans_array = np.full(POLAR_POINT_COUNT, np.nan)
        if self.low_speed_aero:
            self.add_input("cl_low_speed", val=nans_array)
            self.add_output("cd_trim_low_speed", shape=POLAR_POINT_COUNT)
        else:
            self.add_input("data:aerodynamics:aircraft:cruise:CL", val=nans_array)
            self.add_output("data:aerodynamics:aircraft:cruise:CD:trim", shape=POLAR_POINT_COUNT)

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        if self.low_speed_aero:
            cl = inputs["cl_low_speed"]
        else:
            cl = inputs["data:aerodynamics:aircraft:cruise:CL"]

        cd_trim = []

        for cl_val in cl:
            cd_trim.append(5.89 * pow(10, -4) * cl_val)

        if self.low_speed_aero:
            outputs["cd_trim_low_speed"] = cd_trim
        else:
            outputs["data:aerodynamics:aircraft:cruise:CD:trim"] = cd_trim



# -*- coding: utf-8 -*-
"""
  Sellar functions
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from math import exp

from .functions_base import FunctionsBase


class Functions(FunctionsBase):
    """ An OpenMDAO component to encapsulate Functions discipline """

    # pylint: disable=invalid-name
    # pylint: disable=arguments-differ
    def compute(self, inputs, outputs):
        """ Functions computation """

        z2 = inputs["z"][1]
        x1 = inputs["x"]
        y1 = inputs["y1"]
        y2 = inputs["y2"]

        outputs["f"] = x1 ** 2 + z2 + y1 + exp(-y2)
        outputs["g1"] = 3.16 - y1
        outputs["g2"] = y2 - 24.0



#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, Sequence, Optional, Tuple

import numpy as np
from fastoad.constants import FlightPhase
from numpy.testing import assert_allclose

from ..propulsion import EngineSet
from ..propulsion import IPropulsion


class DummyEngine(IPropulsion):
    def __init__(self, max_thrust, max_sfc):
        """
        Dummy engine model.

        Max thrust does not depend on flight conditions.
        SFC is proportional to thrust rate.

        :param max_thrust: thrust when thrust rate = 1.0
        :param max_sfc: SFC when thrust rate = 1.0
        """
        self.max_thrust = max_thrust
        self.max_sfc = max_sfc

    def compute_flight_points(
        self,
        mach: Union[float, Sequence],
        altitude: Union[float, Sequence],
        phase: Union[FlightPhase, Sequence],
        use_thrust_rate: Optional[Union[bool, Sequence]] = None,
        thrust_rate: Optional[Union[float, Sequence]] = None,
        thrust: Optional[Union[float, Sequence]] = None,
    ) -> Tuple[Union[float, Sequence], Union[float, Sequence], Union[float, Sequence]]:

        if use_thrust_rate or thrust is None:
            thrust = self.max_thrust * thrust_rate
        else:
            thrust_rate = thrust / self.max_thrust

        sfc = self.max_sfc * thrust_rate

        return sfc, thrust_rate, thrust


def test_EngineSet():
    """Tests behaviour of EngineSet"""
    engine = DummyEngine(1.2e5, 1.0e-5)

    # input = thrust rate
    thrust_rate = np.linspace(0.0, 1.0, 20)
    for engine_count in [1, 2, 3, 4]:
        engine_set = EngineSet(engine, engine_count)
        sfc, _, thrust = engine_set.compute_flight_points(0.0, 0.0, 1, thrust_rate=thrust_rate)
        assert_allclose(sfc, engine.max_sfc * thrust_rate)
        assert_allclose(thrust, engine.max_thrust * engine_count * thrust_rate)

    # input = thrust
    thrust = np.linspace(0.0, engine.max_thrust, 30)
    for engine_count in [1, 2, 3, 4]:
        engine_set = EngineSet(engine, engine_count)
        sfc, thrust_rate, _ = engine_set.compute_flight_points(0.0, 0.0, 1, thrust=thrust)
        assert_allclose(thrust_rate, thrust / engine_count / engine.max_thrust)
        assert_allclose(sfc, engine.max_sfc * thrust_rate)



"""
test module for modules in aerodynamics/components
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import numpy as np
from fastoad.io import VariableIO
from fastoad.utils.physics import Atmosphere
from openmdao.core.group import Group
from openmdao.core.indepvarcomp import IndepVarComp
from pytest import approx

from tests.testing_utilities import run_system
from ..cd0 import CD0
from ..cd_compressibility import CdCompressibility
from ..cd_trim import CdTrim
from ..compute_low_speed_aero import ComputeAerodynamicsLowSpeed
from ..compute_polar import ComputePolar
from ..compute_reynolds import ComputeReynolds
from ..high_lift_aero import ComputeDeltaHighLift
from ..oswald import OswaldCoefficient


def get_indep_var_comp(var_names):
    """ Reads required input data and returns an IndepVarcomp() instance"""
    reader = VariableIO(pth.join(pth.dirname(__file__), "data", "aerodynamics_inputs.xml"))
    reader.path_separator = ":"
    ivc = reader.read(only=var_names).to_ivc()
    return ivc


def test_high_lift_aero():
    """ Tests ComputeDeltaHighLift """
    input_list = [
        "data:geometry:wing:sweep_0",
        "data:geometry:wing:sweep_100_outer",
        "data:geometry:flap:chord_ratio",
        "data:geometry:flap:span_ratio",
        "data:geometry:slat:chord_ratio",
        "data:geometry:slat:span_ratio",
    ]

    def get_cl_cd(slat_angle, flap_angle, mach, landing_flag):
        ivc = get_indep_var_comp(input_list)
        if landing_flag:
            ivc.add_output("data:mission:sizing:landing:slat_angle", slat_angle, units="deg")
            ivc.add_output("data:mission:sizing:landing:flap_angle", flap_angle, units="deg")
        else:
            ivc.add_output("data:mission:sizing:takeoff:slat_angle", slat_angle, units="deg")
            ivc.add_output("data:mission:sizing:takeoff:flap_angle", flap_angle, units="deg")
        ivc.add_output("data:aerodynamics:aircraft:landing:mach", mach)
        component = ComputeDeltaHighLift()
        component.options["landing_flag"] = landing_flag
        problem = run_system(component, ivc)
        if landing_flag:
            return problem["data:aerodynamics:high_lift_devices:landing:CL"]

        return problem["delta_cl_takeoff"], problem["delta_cd_takeoff"]

    cl, cd = get_cl_cd(18, 10, 0.2, False)
    assert cl == approx(0.516, abs=1e-3)
    assert cd == approx(0.01430, abs=1e-5)

    cl, cd = get_cl_cd(27, 35, 0.4, False)
    assert cl == approx(1.431, abs=1e-3)
    assert cd == approx(0.04644, abs=1e-5)

    cl = get_cl_cd(22, 20, 0.2, True)[0]
    assert cl == approx(0.935, abs=1e-3)


def test_oswald():
    """ Tests OswaldCoefficient """
    input_list = [
        "data:geometry:wing:area",
        "data:geometry:wing:span",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:sweep_25",
    ]

    def get_coeff(mach):
        ivc = get_indep_var_comp(input_list)
        ivc.add_output("data:TLAR:cruise_mach", mach)
        problem = run_system(OswaldCoefficient(), ivc)
        return problem["data:aerodynamics:aircraft:cruise:induced_drag_coefficient"]

    assert get_coeff(0.2) == approx(0.0465, abs=1e-4)
    assert get_coeff(0.8) == approx(0.0530, abs=1e-4)


def test_cd0():
    """ Tests CD0 (test of the group for comparison to legacy) """
    input_list = [
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:wetted_area",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:horizontal_tail:MAC:length",
        "data:geometry:horizontal_tail:sweep_25",
        "data:geometry:horizontal_tail:thickness_ratio",
        "data:geometry:horizontal_tail:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:propulsion:engine:count",
        "data:geometry:propulsion:fan:length",
        "data:geometry:propulsion:nacelle:length",
        "data:geometry:propulsion:nacelle:wetted_area",
        "data:geometry:propulsion:pylon:length",
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:aircraft:wetted_area",
        "data:geometry:vertical_tail:MAC:length",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:vertical_tail:thickness_ratio",
        "data:geometry:vertical_tail:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:wetted_area",
    ]

    def get_cd0(alt, mach, cl):
        reynolds = Atmosphere(alt).get_unitary_reynolds(mach)

        ivc = get_indep_var_comp(input_list)
        ivc.add_output("data:TLAR:cruise_mach", mach)
        ivc.add_output("data:aerodynamics:wing:cruise:reynolds", reynolds)
        ivc.add_output(
            "data:aerodynamics:aircraft:cruise:CL", 150 * [cl]
        )  # needed because size of input array is fixed
        problem = run_system(CD0(), ivc)
        return problem["data:aerodynamics:aircraft:cruise:CD0"][0]

    assert get_cd0(35000, 0.78, 0.5) == approx(0.01975, abs=1e-5)
    assert get_cd0(0, 0.2, 0.9) == approx(0.02727, abs=1e-5)


def test_cd_compressibility():
    """ Tests CdCompressibility """

    def get_cd_compressibility(mach, cl):
        ivc = IndepVarComp()
        ivc.add_output(
            "data:aerodynamics:aircraft:cruise:CL", 150 * [cl]
        )  # needed because size of input array is fixed
        ivc.add_output("data:TLAR:cruise_mach", mach)
        problem = run_system(CdCompressibility(), ivc)
        return problem["data:aerodynamics:aircraft:cruise:CD:compressibility"][0]

    assert get_cd_compressibility(0.78, 0.2) == approx(0.00028, abs=1e-5)
    assert get_cd_compressibility(0.78, 0.35) == approx(0.00028, abs=1e-5)
    assert get_cd_compressibility(0.78, 0.5) == approx(0.00045, abs=1e-5)
    assert get_cd_compressibility(0.84, 0.2) == approx(0.00359, abs=1e-5)
    assert get_cd_compressibility(0.84, 0.35) == approx(0.00359, abs=1e-5)
    assert get_cd_compressibility(0.84, 0.5) == approx(0.00580, abs=1e-5)
    assert get_cd_compressibility(0.2, 0.9) == approx(0.0, abs=1e-5)


def test_cd_trim():
    """ Tests CdTrim """

    def get_cd_trim(cl):
        ivc = IndepVarComp()
        ivc.add_output(
            "data:aerodynamics:aircraft:cruise:CL", 150 * [cl]
        )  # needed because size of input array is fixed
        problem = run_system(CdTrim(), ivc)
        return problem["data:aerodynamics:aircraft:cruise:CD:trim"][0]

    assert get_cd_trim(0.5) == approx(0.0002945, abs=1e-6)
    assert get_cd_trim(0.9) == approx(0.0005301, abs=1e-6)


def test_polar():
    """ Tests ComputePolar """

    # Need to plug Cd modules, Reynolds and Oswald

    input_list = [
        "tuning:aerodynamics:aircraft:cruise:CD:k",
        "tuning:aerodynamics:aircraft:cruise:CD:offset",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:k",
        "tuning:aerodynamics:aircraft:cruise:CD:winglet_effect:offset",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:fuselage:length",
        "data:geometry:fuselage:wetted_area",
        "data:geometry:fuselage:maximum_width",
        "data:geometry:wing:area",
        "data:geometry:horizontal_tail:MAC:length",
        "data:geometry:horizontal_tail:sweep_25",
        "data:geometry:horizontal_tail:thickness_ratio",
        "data:geometry:horizontal_tail:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:propulsion:engine:count",
        "data:geometry:propulsion:fan:length",
        "data:geometry:propulsion:nacelle:length",
        "data:geometry:propulsion:nacelle:wetted_area",
        "data:geometry:propulsion:pylon:length",
        "data:geometry:propulsion:pylon:wetted_area",
        "data:geometry:wing:area",
        "data:geometry:aircraft:wetted_area",
        "data:geometry:vertical_tail:MAC:length",
        "data:geometry:vertical_tail:sweep_25",
        "data:geometry:vertical_tail:thickness_ratio",
        "data:geometry:vertical_tail:wetted_area",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:thickness_ratio",
        "data:geometry:wing:wetted_area",
        "data:geometry:wing:span",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:tip:chord",
        "data:TLAR:cruise_mach",
        "data:mission:sizing:cruise:altitude",
    ]
    group = Group()
    group.add_subsystem("reynolds", ComputeReynolds(), promotes=["*"])
    group.add_subsystem("oswald", OswaldCoefficient(), promotes=["*"])
    group.add_subsystem("cd0", CD0(), promotes=["*"])
    group.add_subsystem("cd_compressibility", CdCompressibility(), promotes=["*"])
    group.add_subsystem("cd_trim", CdTrim(), promotes=["*"])
    group.add_subsystem("polar", ComputePolar(), promotes=["*"])

    ivc = get_indep_var_comp(input_list)
    ivc.add_output("data:aerodynamics:aircraft:cruise:CL", np.arange(0.0, 1.5, 0.01))

    problem = run_system(group, ivc)

    cd = problem["data:aerodynamics:aircraft:cruise:CD"]
    cl = problem["data:aerodynamics:aircraft:cruise:CL"]

    assert cd[cl == 0.0] == approx(0.02030, abs=1e-5)
    assert cd[cl == 0.2] == approx(0.02209, abs=1e-5)
    assert cd[cl == 0.42] == approx(0.02897, abs=1e-5)
    assert cd[cl == 0.85] == approx(0.11781, abs=1e-5)

    assert problem["data:aerodynamics:aircraft:cruise:optimal_CL"] == approx(0.54, abs=1e-3)
    assert problem["data:aerodynamics:aircraft:cruise:optimal_CD"] == approx(0.03550, abs=1e-5)


def test_low_speed_aero():
    """ Tests group ComputeAerodynamicsLowSpeed """
    input_list = [
        "data:geometry:fuselage:maximum_width",
        "data:geometry:fuselage:maximum_height",
        "data:geometry:wing:span",
        "data:geometry:wing:aspect_ratio",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:sweep_25",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:area",
        "data:geometry:wing:tip:thickness_ratio",
    ]
    ivc = get_indep_var_comp(input_list)

    problem = run_system(ComputeAerodynamicsLowSpeed(), ivc)

    assert problem["data:aerodynamics:aircraft:takeoff:CL_alpha"] == approx(5.0, abs=1e-1)



"""
Test module for geometry functions of cg components
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import pytest
from fastoad.io import VariableIO

from tests.testing_utilities import run_system
from ..cg import ComputeAircraftCG
from ..cg_components import ComputeCGLoadCase1
from ..cg_components import ComputeCGLoadCase2
from ..cg_components import ComputeCGLoadCase3
from ..cg_components import ComputeCGLoadCase4
from ..cg_components import ComputeCGRatioAft
from ..cg_components import ComputeControlSurfacesCG
from ..cg_components import ComputeGlobalCG
from ..cg_components import ComputeMaxCGratio
from ..cg_components import ComputeOthersCG
from ..cg_components import ComputeTanksCG
from ..cg_components import ComputeWingCG


# pylint: disable=redefined-outer-name  # needed for pytest fixtures
@pytest.fixture(scope="module")
def input_xml() -> VariableIO:
    """
    :return: access to the sample xml data
    """
    # TODO: have more consistency in input data (no need for the whole geometry_inputs_full.xml)
    return VariableIO(pth.join(pth.dirname(__file__), "data", "cg_inputs.xml"))


def test_compute_cg_control_surfaces(input_xml):
    """ Tests computation of control surfaces center of gravity """

    input_list = [
        "data:geometry:wing:MAC:leading_edge:x:local",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:y",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:MAC:at25percent:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeControlSurfacesCG(), input_vars)

    x_cg_a4 = problem["data:weight:airframe:flight_controls:CG:x"]
    assert x_cg_a4 == pytest.approx(19.24, abs=1e-2)


def test_compute_cg_loadcase1(input_xml):
    """ Tests computation of center of gravity for load case 1 """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:weight:payload:PAX:CG:x",
        "data:weight:payload:rear_fret:CG:x",
        "data:weight:payload:front_fret:CG:x",
        "data:TLAR:NPAX",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft_empty:CG:x", 699570.01 / 40979.11)
    input_vars.add_output("data:weight:aircraft_empty:mass", 40979.11)

    problem = run_system(ComputeCGLoadCase1(), input_vars)

    cg_ratio_lc1 = problem["data:weight:aircraft:load_case_1:CG:MAC_position"]
    assert cg_ratio_lc1 == pytest.approx(0.364924, abs=1e-6)


def test_compute_cg_loadcase2(input_xml):
    """ Tests computation of center of gravity for load case 2 """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:weight:payload:PAX:CG:x",
        "data:weight:payload:rear_fret:CG:x",
        "data:weight:payload:front_fret:CG:x",
        "data:TLAR:NPAX",
        "data:weight:aircraft:MFW",
        "data:weight:fuel_tank:CG:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft_empty:CG:x", 699570.01 / 40979.11)
    input_vars.add_output("data:weight:aircraft_empty:mass", 40979.11)

    problem = run_system(ComputeCGLoadCase2(), input_vars)

    cg_ratio_lc2 = problem["data:weight:aircraft:load_case_2:CG:MAC_position"]
    assert cg_ratio_lc2 == pytest.approx(0.285139, abs=1e-6)


def test_compute_cg_loadcase3(input_xml):
    """ Tests computation of center of gravity for load case 3 """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:weight:payload:PAX:CG:x",
        "data:weight:payload:rear_fret:CG:x",
        "data:weight:payload:front_fret:CG:x",
        "data:TLAR:NPAX",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft_empty:CG:x", 699570.01 / 40979.11)
    input_vars.add_output("data:weight:aircraft_empty:mass", 40979.11)

    problem = run_system(ComputeCGLoadCase3(), input_vars)

    cg_ratio_lc3 = problem["data:weight:aircraft:load_case_3:CG:MAC_position"]
    assert cg_ratio_lc3 == pytest.approx(0.386260, abs=1e-6)


def test_compute_cg_loadcase4(input_xml):
    """ Tests computation of center of gravity for load case 4 """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:weight:payload:PAX:CG:x",
        "data:weight:payload:rear_fret:CG:x",
        "data:weight:payload:front_fret:CG:x",
        "data:TLAR:NPAX",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft_empty:CG:x", 699570.01 / 40979.11)
    input_vars.add_output("data:weight:aircraft_empty:mass", 40979.11)

    problem = run_system(ComputeCGLoadCase4(), input_vars)

    cg_ratio_lc4 = problem["data:weight:aircraft:load_case_4:CG:MAC_position"]
    assert cg_ratio_lc4 == pytest.approx(0.388971, abs=1e-6)


def test_compute_cg_others(input_xml):
    """ Tests computation of other components center of gravity """

    input_list = [
        "data:geometry:wing:MAC:leading_edge:x:local",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:root:chord",
        "data:geometry:fuselage:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:fuselage:front_length",
        "data:geometry:fuselage:rear_length",
        "data:weight:propulsion:engine:CG:x",
        "data:weight:furniture:passenger_seats:CG:x",
        "data:weight:propulsion:engine:mass",
        "data:geometry:cabin:NPAX1",
        "data:geometry:cabin:seats:economical:count_by_row",
        "data:geometry:cabin:seats:economical:length",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeOthersCG(), input_vars)

    x_cg_a2 = problem["data:weight:airframe:fuselage:CG:x"]
    assert x_cg_a2 == pytest.approx(16.88, abs=1e-2)
    x_cg_a52 = problem["data:weight:airframe:landing_gear:front:CG:x"]
    assert x_cg_a52 == pytest.approx(5.18, abs=1e-2)
    x_cg_a6 = problem["data:weight:airframe:pylon:CG:x"]
    assert x_cg_a6 == pytest.approx(13.5, abs=1e-1)
    x_cg_a7 = problem["data:weight:airframe:paint:CG:x"]
    assert x_cg_a7 == pytest.approx(0.0, abs=1e-1)

    x_cg_b2 = problem["data:weight:propulsion:fuel_lines:CG:x"]
    assert x_cg_b2 == pytest.approx(13.5, abs=1e-1)
    x_cg_b3 = problem["data:weight:propulsion:unconsumables:CG:x"]
    assert x_cg_b3 == pytest.approx(13.5, abs=1e-1)

    x_cg_c11 = problem["data:weight:systems:power:auxiliary_power_unit:CG:x"]
    assert x_cg_c11 == pytest.approx(35.63, abs=1e-2)
    x_cg_c12 = problem["data:weight:systems:power:electric_systems:CG:x"]
    assert x_cg_c12 == pytest.approx(18.75, abs=1e-2)
    x_cg_c13 = problem["data:weight:systems:power:hydraulic_systems:CG:x"]
    assert x_cg_c13 == pytest.approx(18.75, abs=1e-2)
    x_cg_c21 = problem["data:weight:systems:life_support:insulation:CG:x"]
    assert x_cg_c21 == pytest.approx(16.88, abs=1e-2)
    x_cg_c22 = problem["data:weight:systems:life_support:air_conditioning:CG:x"]
    assert x_cg_c22 == pytest.approx(16.62, abs=1e-2)
    x_cg_c23 = problem["data:weight:systems:life_support:de-icing:CG:x"]
    assert x_cg_c23 == pytest.approx(15.79, abs=1e-2)
    x_cg_c24 = problem["data:weight:systems:life_support:cabin_lighting:CG:x"]
    assert x_cg_c24 == pytest.approx(16.88, abs=1e-2)
    x_cg_c25 = problem["data:weight:systems:life_support:seats_crew_accommodation:CG:x"]
    assert x_cg_c25 == pytest.approx(16.62, abs=1e-2)
    x_cg_c26 = problem["data:weight:systems:life_support:oxygen:CG:x"]
    assert x_cg_c26 == pytest.approx(16.62, abs=1e-2)
    x_cg_c27 = problem["data:weight:systems:life_support:safety_equipment:CG:x"]
    assert x_cg_c27 == pytest.approx(16.1, abs=1e-1)
    x_cg_c3 = problem["data:weight:systems:navigation:CG:x"]
    assert x_cg_c3 == pytest.approx(5.52, abs=1e-2)
    x_cg_c4 = problem["data:weight:systems:transmission:CG:x"]
    assert x_cg_c4 == pytest.approx(18.75, abs=1e-2)
    x_cg_c51 = problem["data:weight:systems:operational:radar:CG:x"]
    assert x_cg_c51 == pytest.approx(0.75, abs=1e-2)
    x_cg_c52 = problem["data:weight:systems:operational:cargo_hold:CG:x"]
    assert x_cg_c52 == pytest.approx(16.62, abs=1e-2)

    x_cg_d3 = problem["data:weight:furniture:food_water:CG:x"]
    assert x_cg_d3 == pytest.approx(29.4, abs=1e-1)
    x_cg_d4 = problem["data:weight:furniture:security_kit:CG:x"]
    assert x_cg_d4 == pytest.approx(16.62, abs=1e-2)
    x_cg_d5 = problem["data:weight:furniture:toilets:CG:x"]
    assert x_cg_d5 == pytest.approx(16.62, abs=1e-2)
    x_cg_pl = problem["data:weight:payload:PAX:CG:x"]
    assert x_cg_pl == pytest.approx(16.62, abs=1e-2)
    x_cg_rear_fret = problem["data:weight:payload:rear_fret:CG:x"]
    assert x_cg_rear_fret == pytest.approx(20.87, abs=1e-2)
    x_cg_front_fret = problem["data:weight:payload:front_fret:CG:x"]
    assert x_cg_front_fret == pytest.approx(9.94, abs=1e-2)


def test_compute_cg_ratio_aft(input_xml):
    """ Tests computation of center of gravity with aft estimation """

    input_list = [
        "data:weight:airframe:wing:CG:x",
        "data:weight:airframe:fuselage:CG:x",
        "data:weight:airframe:horizontal_tail:CG:x",
        "data:weight:airframe:vertical_tail:CG:x",
        "data:weight:airframe:flight_controls:CG:x",
        "data:weight:airframe:landing_gear:main:CG:x",
        "data:weight:airframe:landing_gear:front:CG:x",
        "data:weight:airframe:pylon:CG:x",
        "data:weight:airframe:paint:CG:x",
        "data:weight:airframe:wing:mass",
        "data:weight:airframe:fuselage:mass",
        "data:weight:airframe:horizontal_tail:mass",
        "data:weight:airframe:vertical_tail:mass",
        "data:weight:airframe:flight_controls:mass",
        "data:weight:airframe:landing_gear:main:mass",
        "data:weight:airframe:landing_gear:front:mass",
        "data:weight:airframe:pylon:mass",
        "data:weight:airframe:paint:mass",
        "data:weight:propulsion:engine:CG:x",
        "data:weight:propulsion:fuel_lines:CG:x",
        "data:weight:propulsion:unconsumables:CG:x",
        "data:weight:propulsion:engine:mass",
        "data:weight:propulsion:fuel_lines:mass",
        "data:weight:propulsion:unconsumables:mass",
        "data:weight:systems:power:auxiliary_power_unit:CG:x",
        "data:weight:systems:power:electric_systems:CG:x",
        "data:weight:systems:power:hydraulic_systems:CG:x",
        "data:weight:systems:life_support:insulation:CG:x",
        "data:weight:systems:life_support:air_conditioning:CG:x",
        "data:weight:systems:life_support:de-icing:CG:x",
        "data:weight:systems:life_support:cabin_lighting:CG:x",
        "data:weight:systems:life_support:seats_crew_accommodation:CG:x",
        "data:weight:systems:life_support:oxygen:CG:x",
        "data:weight:systems:life_support:safety_equipment:CG:x",
        "data:weight:systems:navigation:CG:x",
        "data:weight:systems:transmission:CG:x",
        "data:weight:systems:operational:radar:CG:x",
        "data:weight:systems:operational:cargo_hold:CG:x",
        "data:weight:systems:flight_kit:CG:x",
        "data:weight:systems:power:auxiliary_power_unit:mass",
        "data:weight:systems:power:electric_systems:mass",
        "data:weight:systems:power:hydraulic_systems:mass",
        "data:weight:systems:life_support:insulation:mass",
        "data:weight:systems:life_support:air_conditioning:mass",
        "data:weight:systems:life_support:de-icing:mass",
        "data:weight:systems:life_support:cabin_lighting:mass",
        "data:weight:systems:life_support:seats_crew_accommodation:mass",
        "data:weight:systems:life_support:oxygen:mass",
        "data:weight:systems:life_support:safety_equipment:mass",
        "data:weight:systems:navigation:mass",
        "data:weight:systems:transmission:mass",
        "data:weight:systems:operational:radar:mass",
        "data:weight:systems:operational:cargo_hold:mass",
        "data:weight:systems:flight_kit:mass",
        "data:weight:furniture:passenger_seats:CG:x",
        "data:weight:furniture:food_water:CG:x",
        "data:weight:furniture:security_kit:CG:x",
        "data:weight:furniture:toilets:CG:x",
        "data:weight:furniture:passenger_seats:mass",
        "data:weight:furniture:food_water:mass",
        "data:weight:furniture:security_kit:mass",
        "data:weight:furniture:toilets:mass",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeCGRatioAft(), input_vars)

    empty_mass = problem["data:weight:aircraft_empty:mass"]
    assert empty_mass == pytest.approx(41120, abs=11)
    cg_ratio_aft = problem["data:weight:aircraft:empty:CG:MAC_position"]
    assert cg_ratio_aft == pytest.approx(0.374702, abs=1e-6)


def test_compute_cg_tanks(input_xml):
    """ Tests computation of tanks center of gravity """

    input_list = [
        "data:geometry:wing:spar_ratio:front:root",
        "data:geometry:wing:spar_ratio:front:kink",
        "data:geometry:wing:spar_ratio:front:tip",
        "data:geometry:wing:spar_ratio:rear:root",
        "data:geometry:wing:spar_ratio:rear:kink",
        "data:geometry:wing:spar_ratio:rear:tip",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:leading_edge:x:local",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:tip:leading_edge:x:local",
        "data:geometry:wing:MAC:at25percent:x",
        "data:geometry:fuselage:maximum_width",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeTanksCG(), input_vars)

    x_cg_tank = problem["data:weight:fuel_tank:CG:x"]
    assert x_cg_tank == pytest.approx(16.05, abs=1e-2)


def test_compute_cg_wing(input_xml):
    """ Tests computation of wing center of gravity """

    input_list = [
        "data:geometry:wing:kink:span_ratio",
        "data:geometry:wing:spar_ratio:front:root",
        "data:geometry:wing:spar_ratio:front:kink",
        "data:geometry:wing:spar_ratio:front:tip",
        "data:geometry:wing:spar_ratio:rear:root",
        "data:geometry:wing:spar_ratio:rear:kink",
        "data:geometry:wing:spar_ratio:rear:tip",
        "data:geometry:wing:span",
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:leading_edge:x:local",
        "data:geometry:wing:root:chord",
        "data:geometry:wing:kink:chord",
        "data:geometry:wing:tip:chord",
        "data:geometry:wing:root:y",
        "data:geometry:wing:kink:leading_edge:x:local",
        "data:geometry:wing:kink:y",
        "data:geometry:wing:tip:y",
        "data:geometry:wing:tip:leading_edge:x:local",
        "data:geometry:wing:MAC:at25percent:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeWingCG(), input_vars)

    x_cg_wing = problem["data:weight:airframe:wing:CG:x"]
    assert x_cg_wing == pytest.approx(16.67, abs=1e-2)


def test_compute_global_cg(input_xml):
    """ Tests computation of global center of gravity """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
        "data:weight:payload:PAX:CG:x",
        "data:weight:payload:rear_fret:CG:x",
        "data:weight:payload:front_fret:CG:x",
        "data:TLAR:NPAX",
        "data:weight:aircraft:MFW",
        "data:weight:fuel_tank:CG:x",
        "data:weight:airframe:wing:CG:x",
        "data:weight:airframe:fuselage:CG:x",
        "data:weight:airframe:horizontal_tail:CG:x",
        "data:weight:airframe:vertical_tail:CG:x",
        "data:weight:airframe:flight_controls:CG:x",
        "data:weight:airframe:landing_gear:main:CG:x",
        "data:weight:airframe:landing_gear:front:CG:x",
        "data:weight:airframe:pylon:CG:x",
        "data:weight:airframe:paint:CG:x",
        "data:weight:airframe:wing:mass",
        "data:weight:airframe:fuselage:mass",
        "data:weight:airframe:horizontal_tail:mass",
        "data:weight:airframe:vertical_tail:mass",
        "data:weight:airframe:flight_controls:mass",
        "data:weight:airframe:landing_gear:main:mass",
        "data:weight:airframe:landing_gear:front:mass",
        "data:weight:airframe:pylon:mass",
        "data:weight:airframe:paint:mass",
        "data:weight:propulsion:engine:CG:x",
        "data:weight:propulsion:fuel_lines:CG:x",
        "data:weight:propulsion:unconsumables:CG:x",
        "data:weight:propulsion:engine:mass",
        "data:weight:propulsion:fuel_lines:mass",
        "data:weight:propulsion:unconsumables:mass",
        "data:weight:systems:power:auxiliary_power_unit:CG:x",
        "data:weight:systems:power:electric_systems:CG:x",
        "data:weight:systems:power:hydraulic_systems:CG:x",
        "data:weight:systems:life_support:insulation:CG:x",
        "data:weight:systems:life_support:air_conditioning:CG:x",
        "data:weight:systems:life_support:de-icing:CG:x",
        "data:weight:systems:life_support:cabin_lighting:CG:x",
        "data:weight:systems:life_support:seats_crew_accommodation:CG:x",
        "data:weight:systems:life_support:oxygen:CG:x",
        "data:weight:systems:life_support:safety_equipment:CG:x",
        "data:weight:systems:navigation:CG:x",
        "data:weight:systems:transmission:CG:x",
        "data:weight:systems:operational:radar:CG:x",
        "data:weight:systems:operational:cargo_hold:CG:x",
        "data:weight:systems:flight_kit:CG:x",
        "data:weight:systems:power:auxiliary_power_unit:mass",
        "data:weight:systems:power:electric_systems:mass",
        "data:weight:systems:power:hydraulic_systems:mass",
        "data:weight:systems:life_support:insulation:mass",
        "data:weight:systems:life_support:air_conditioning:mass",
        "data:weight:systems:life_support:de-icing:mass",
        "data:weight:systems:life_support:cabin_lighting:mass",
        "data:weight:systems:life_support:seats_crew_accommodation:mass",
        "data:weight:systems:life_support:oxygen:mass",
        "data:weight:systems:life_support:safety_equipment:mass",
        "data:weight:systems:navigation:mass",
        "data:weight:systems:transmission:mass",
        "data:weight:systems:operational:radar:mass",
        "data:weight:systems:operational:cargo_hold:mass",
        "data:weight:systems:flight_kit:mass",
        "data:weight:furniture:passenger_seats:CG:x",
        "data:weight:furniture:food_water:CG:x",
        "data:weight:furniture:security_kit:CG:x",
        "data:weight:furniture:toilets:CG:x",
        "data:weight:furniture:passenger_seats:mass",
        "data:weight:furniture:food_water:mass",
        "data:weight:furniture:security_kit:mass",
        "data:weight:furniture:toilets:mass",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()

    problem = run_system(ComputeGlobalCG(), input_vars)

    cg_ratio = problem["data:weight:aircraft:CG:aft:MAC_position"]
    assert cg_ratio == pytest.approx(0.430052, abs=1e-6)


def test_compute_max_cg_ratio(input_xml):
    """ Tests computation of maximum center of gravity ratio """

    input_list = []

    input_vars = input_xml.read(only=input_list).to_ivc()

    input_vars.add_output("data:weight:aircraft:empty:CG:MAC_position", 0.387846)
    input_vars.add_output("data:weight:aircraft:load_case_1:CG:MAC_position", 0.364924)
    input_vars.add_output("data:weight:aircraft:load_case_2:CG:MAC_position", 0.285139)
    input_vars.add_output("data:weight:aircraft:load_case_3:CG:MAC_position", 0.386260)
    input_vars.add_output("data:weight:aircraft:load_case_4:CG:MAC_position", 0.388971)

    problem = run_system(ComputeMaxCGratio(), input_vars)

    cg_ratio = problem["data:weight:aircraft:CG:aft:MAC_position"]
    assert cg_ratio == pytest.approx(0.438971, abs=1e-6)


def test_compute_aircraft_cg(input_xml):
    """ Tests computation of static margin """

    input_list = [
        "data:geometry:wing:MAC:length",
        "data:geometry:wing:MAC:at25percent:x",
    ]

    input_vars = input_xml.read(only=input_list).to_ivc()
    input_vars.add_output("data:weight:aircraft:CG:aft:MAC_position", 0.388971)

    problem = run_system(ComputeAircraftCG(), input_vars)

    cg_global = problem["data:weight:aircraft:CG:aft:x"]
    assert cg_global == pytest.approx(17.1, abs=1e-1)



"""Module for registering services."""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.models.propulsion import IOMPropulsionWrapper
from fastoad.module_management.constants import SERVICE_PROPULSION_WRAPPER
from fastoad.module_management.exceptions import FastIncompatibleServiceClass
from pelix.ipopo.decorators import ComponentFactory, Provides


class RegisterService:
    def __init__(self, provider_id: str, service_id: str, base_class: type = None):
        """
        Decorator class that allows to register a service.

        The service can be associated to a base class (or interface). Then all registered
        classes must inherit from this base class.

        Warning: the module must be started as an iPOPO bundle for the decorator to work.

        :param provider_id: the identifier of the service provider to register
        :param service_id: the identifier of the service
        :param base_class: the class that should be parent to the registered class
        """
        self._id = provider_id
        self._service_id = service_id
        self._base_class = base_class

    def __call__(self, service_class: type) -> type:
        if self._base_class and not issubclass(service_class, self._base_class):
            raise FastIncompatibleServiceClass(service_class, self._service_id, self._base_class)

        return ComponentFactory(self._id)(Provides(self._service_id)(service_class))


class RegisterPropulsion(RegisterService):
    def __init__(self, model_id):
        """
        Decorator class for registering a propulsion-dedicated model.

        Warning: the module must be started as an iPOPO bundle for the decorator to work.

        :param model_id: the identifier of the propulsion model
        """
        super().__init__(model_id, SERVICE_PROPULSION_WRAPPER, IOMPropulsionWrapper)



"""
    Estimation of control surfaces center of gravity
"""

#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import numpy as np
from openmdao.core.explicitcomponent import ExplicitComponent


class ComputeControlSurfacesCG(ExplicitComponent):
    # TODO: Document equations. Cite sources
    """ Control surfaces center of gravity estimation """

    def setup(self):
        self.add_input("data:geometry:wing:MAC:length", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:chord", val=np.nan, units="m")
        self.add_input("data:geometry:wing:root:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:leading_edge:x:local", val=np.nan, units="m")
        self.add_input("data:geometry:wing:kink:y", val=np.nan, units="m")
        self.add_input("data:geometry:wing:MAC:at25percent:x", val=np.nan, units="m")

        self.add_output("data:weight:airframe:flight_controls:CG:x", units="m")

        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        l0_wing = inputs["data:geometry:wing:MAC:length"]
        x0_wing = inputs["data:geometry:wing:MAC:leading_edge:x:local"]
        y0_wing = inputs["data:geometry:wing:MAC:y"]
        l2_wing = inputs["data:geometry:wing:root:chord"]
        l3_wing = inputs["data:geometry:wing:kink:chord"]
        y2_wing = inputs["data:geometry:wing:root:y"]
        x3_wing = inputs["data:geometry:wing:kink:leading_edge:x:local"]
        y3_wing = inputs["data:geometry:wing:kink:y"]
        fa_length = inputs["data:geometry:wing:MAC:at25percent:x"]

        # TODO: build generic functions to estimate the chord, leading edge,
        # trailing edge with respect to span wise position
        x_leading_edge = x3_wing * (y0_wing - y2_wing) / (y3_wing - y2_wing)
        l_cg_control = l2_wing + (y0_wing - y2_wing) / (y3_wing - y2_wing) * (l3_wing - l2_wing)
        x_cg_control = x_leading_edge + l_cg_control
        x_cg_control_absolute = fa_length - 0.25 * l0_wing - x0_wing + x_cg_control

        outputs["data:weight:airframe:flight_controls:CG:x"] = x_cg_control_absolute



"""
test module for wing area computation
"""
#  This file is part of FAST : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2020  ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import openmdao.api as om
from numpy.testing import assert_allclose

from tests.testing_utilities import run_system
from ..compute_wing_area import ComputeWingArea

DATA_FOLDER_PATH = pth.join(pth.dirname(__file__), "data")


def test_compute_wing_area():
    # Driven by fuel
    ivc = om.IndepVarComp()
    ivc.add_output("data:geometry:wing:aspect_ratio", 9.48)
    ivc.add_output("data:geometry:wing:root:thickness_ratio", 0.15)
    ivc.add_output("data:geometry:wing:tip:thickness_ratio", 0.11)
    ivc.add_output("data:mission:sizing:fuel", val=20500, units="kg")
    ivc.add_output("data:TLAR:approach_speed", val=132, units="kn")
    ivc.add_output("data:weight:aircraft:MLW", val=66300, units="kg")

    ivc.add_output("data:weight:aircraft:MFW", val=21000, units="kg")
    ivc.add_output("data:aerodynamics:aircraft:landing:CL_max", val=2.80)

    problem = run_system(ComputeWingArea(), ivc)
    assert_allclose(problem["data:geometry:wing:area"], 133.97, atol=1e-2)
    assert_allclose(
        problem["data:aerodynamics:aircraft:landing:additional_CL_capacity"], 0.199, atol=1e-2
    )
    assert_allclose(problem["data:weight:aircraft:additional_fuel_capacity"], 500.0, atol=1.0)

    # Driven by CL max
    ivc = om.IndepVarComp()
    ivc.add_output("data:geometry:wing:aspect_ratio", 9.48)
    ivc.add_output("data:geometry:wing:root:thickness_ratio", 0.15)
    ivc.add_output("data:geometry:wing:tip:thickness_ratio", 0.11)
    ivc.add_output("data:mission:sizing:fuel", val=15000, units="kg")
    ivc.add_output("data:TLAR:approach_speed", val=132, units="kn")
    ivc.add_output("data:weight:aircraft:MLW", val=66300, units="kg")

    ivc.add_output("data:weight:aircraft:MFW", val=21000, units="kg")
    ivc.add_output("data:aerodynamics:aircraft:landing:CL_max", val=2.80)

    problem = run_system(ComputeWingArea(), ivc)
    assert_allclose(problem["data:geometry:wing:area"], 124.38, atol=1e-2)
    assert_allclose(
        problem["data:aerodynamics:aircraft:landing:additional_CL_capacity"], 0.0, atol=1e-2
    )
    assert_allclose(problem["data:weight:aircraft:additional_fuel_capacity"], 6000.0, atol=1.0)
